import pytest

from starlette.datastructures import FormData

from bocadillo import App
from bocadillo.constants import ALL_HTTP_METHODS


@pytest.mark.parametrize("method", ALL_HTTP_METHODS)
def test_method(app: App, client, method: str):
    req_method = None

    @app.route("/")
    class Index:
        async def handle(self, req, res):
            nonlocal req_method
            req_method = req.method

    r = getattr(client, method.lower())("/")
    assert req_method == method


@pytest.mark.parametrize(
    "attr, value",
    [
        (None, "http://testserver/foo?bar=err"),
        ("path", "/foo"),
        ("scheme", "http"),
        ("hostname", "testserver"),
        ("query", "bar=err"),
        ("is_secure", False),
    ],
)
def test_url(app: App, client, attr, value):
    url = None

    @app.route("/foo")
    async def index(req, res):
        nonlocal url
        url = req.url

    client.get("/foo", params={"bar": "err"})
    assert url is not None

    if attr is None:
        assert url == value
    else:
        assert getattr(url, attr) == value


def test_headers(app: App, client):
    @app.route("/")
    async def index(req, res):
        # Key access.
        assert req.headers["x-foo"] == "foo"
        with pytest.raises(KeyError):
            req.headers["x-bar"]

        # Case insensitivity.
        assert req.headers["x-foo"] == req.headers["X-Foo"] == "foo"

        # Defaults.
        assert req.headers.get("x-bar", 1) == 1

    r = client.get("/", headers={"X-Foo": "foo"})
    assert r.status_code == 200


def test_query_params(app: App, client):
    @app.route("/")
    async def index(req, res):
        assert req.url.query == "a=alpha&b=beta&b=BETA&b=Beta"

        # Key access.
        assert req.query_params["a"] == "alpha"
        with pytest.raises(KeyError):
            req.query_params["c"]

        # Defaults
        assert req.query_params.get("a") == "alpha"
        assert req.query_params.get("c") is None

        # Multiple values
        last_item = "Beta"
        assert req.query_params["b"] == last_item
        assert req.query_params.getlist("b") == ["beta", "BETA", "Beta"]

    r = client.get("/", params={"a": "alpha", "b": ["beta", "BETA", "Beta"]})
    assert r.status_code == 200


def test_raw_body(app: App, client):
    @app.route("/")
    async def index(req, res):
        body = await req.body()
        assert isinstance(body, bytes)
        res.content = body

    r = client.get("/", data="hello")
    assert r.status_code == 200
    assert r.content == b"hello"


def test_form_body(app: App, client):
    @app.route("/")
    async def index(req, res):
        form = await req.form()
        assert isinstance(form, FormData)
        # not actually a dict, so not immediately JSON-serializable
        res.json = dict(form)

    r = client.get("/", data={"foo": "bar"})
    assert r.status_code == 200
    assert r.json() == {"foo": "bar"}


@pytest.mark.parametrize("data, status", [("{", 400), ("{}", 200)])
def test_json_body(app: App, client, data: str, status: int):
    @app.route("/")
    class Index:
        async def post(self, req, res):
            res.json = await req.json()

    assert client.post("/", data=data).status_code == status


@pytest.mark.parametrize(
    "get_stream", [lambda req: req, lambda req: req.stream()]
)
def test_stream_request(app: App, client, get_stream):
    @app.route("/")
    class Index:
        async def get(self, req, res):
            chunks = [
                chunk.decode() async for chunk in get_stream(req) if chunk
            ]
            res.json = chunks

    # For testing, we use a chunk-encoded request. See:
    # http://docs.python-requests.org/en/master/user/advanced/#chunk-encoded-requests

    message = "Hello, world!"

    def stream():
        for _ in range(3):
            yield message

    response = client.get("/", data=stream())
    assert response.json() == [message] * 3



import typing

if typing.TYPE_CHECKING:
    from .applications import App


class SettingsError(Exception):
    """Raised when a setting is missing, ill-declared or invalid."""


class Settings:
    def __init__(self, obj: typing.Optional[typing.Any]):
        for setting in dir(obj):
            if not setting.isupper():
                continue
            value = getattr(obj, setting)
            setattr(self, setting, value)


class LazySettings:
    """A lazy proxy for application settings.

    Once configured, an instance of this class can be used to access settings
    from anywhere in the application code base.

    Such an instance is in fact exposed as `bocadillo.settings`.

    Settings can be accessed using:

    - Dot notation: `settings.FOO`.
    - The `getattr` builtin: `getattr(settings, "FOO")`.
    - The dict-like `.get()` method: `settings.get("FOO", "foo")`.
    """

    def __init__(self):
        self._wrapped = None

    def configure(self, obj: typing.Any = None, **options):
        if self.configured:
            raise RuntimeError("Settings are already configured")

        wrapped = Settings(obj)

        for name, option in options.items():
            if name.startswith("_"):
                continue

            if not name.isupper():
                raise SettingsError(f"Setting {name} must be uppercase.")

            setattr(wrapped, name, option)

        self._wrapped = wrapped

    @property
    def configured(self) -> bool:
        return self._wrapped is not None

    def __getattr__(self, name: str) -> typing.Any:
        if not self.configured:
            raise SettingsError(
                f"Requested setting {name} but settings aren't configured yet."
            )

        value = getattr(self._wrapped, name)

        self.__dict__[name] = value  # cache setting

        return value

    def __setattr__(self, name: str, value: typing.Any):
        if name == "_wrapped":
            self.__dict__.clear()
        else:
            self.__dict__.pop(name, None)  # remove from cache
        super().__setattr__(name, value)

    def get(self, name: str, default: typing.Any = None) -> typing.Any:
        return getattr(self, name, default)

    def _clear(self):
        self._wrapped = None


settings = LazySettings()  # pylint: disable=invalid-name


def configure(app: "App", settings_obj: typing.Any = None, **kwargs) -> "App":
    """Configure the application settings and setup plugins.

    # Parameters
    settings_obj (any): an optional settings object or module.
    **kwargs (any): arbitrary settings, case-insensitive.
    """
    from .plugins import get_plugins

    kwargs = {key.upper(): value for key, value in kwargs.items()}

    settings.configure(settings_obj, **kwargs)

    for plugin in get_plugins().values():
        plugin(app)

    return app  # for convenience



import pytest

from bocadillo import configure, create_client, static

FILE_DIR = "js"
FILE_NAME = "foo.js"
FILE_CONTENTS = "console.log('foo!');"


def _create_asset(static_dir):
    asset = static_dir.mkdir(FILE_DIR).join(FILE_NAME)
    asset.write(FILE_CONTENTS)
    return asset


def test_assets_are_served_at_static_by_default(raw_app, tmpdir_factory):
    static_dir = tmpdir_factory.mktemp("static")
    _create_asset(static_dir)

    app = configure(raw_app, static_dir=str(static_dir))
    client = create_client(app)

    response = client.get(f"/static/{FILE_DIR}/{FILE_NAME}")
    assert response.status_code == 200
    assert response.text == FILE_CONTENTS


def test_if_asset_does_not_exist_then_404(client):
    assert client.get(f"/static/{FILE_DIR}/{FILE_NAME}").status_code == 404


def test_customize_static_root(raw_app, tmpdir_factory):
    static_dir = tmpdir_factory.mktemp("static")
    _create_asset(static_dir)

    app = configure(raw_app, static_dir=str(static_dir), static_root="assets")
    client = create_client(app)

    assert client.get(f"/static/{FILE_DIR}/{FILE_NAME}").status_code == 404
    response = client.get(f"/assets/{FILE_DIR}/{FILE_NAME}")
    assert response.status_code == 200
    assert response.text == FILE_CONTENTS


def test_if_static_dir_is_none_then_no_assets_served(raw_app, tmpdir_factory):
    static_dir = tmpdir_factory.mktemp("static")
    _create_asset(static_dir)

    app = configure(raw_app, static_dir=None)
    client = create_client(app)

    assert client.get(f"/static/{FILE_DIR}/{FILE_NAME}").status_code == 404


def test_mount_extra_static_files_dirs(raw_app, tmpdir_factory):
    static_dir = tmpdir_factory.mktemp("staticfiles")
    _create_asset(static_dir)

    app = configure(raw_app, static_dir=None)
    app.mount("assets", static(str(static_dir)))
    client = create_client(app)

    response = client.get(f"/assets/{FILE_DIR}/{FILE_NAME}")
    assert response.status_code == 200
    assert response.text == FILE_CONTENTS


def test_if_static_dir_does_not_exist_then_no_files_mounted(raw_app):
    with pytest.warns(None) as record:
        configure(raw_app, static_dir="foo")
    assert len(record) == 0


def test_whitenoise_config(raw_app):
    app = configure(
        raw_app, static_root="static", static_config={"max_age": 30}
    )
    whitenoise = app._children["/static"]
    assert whitenoise.max_age == 30



import pytest

from bocadillo import App, ExpectedAsync, hooks

from .utils import class_hooks, function_hooks


@pytest.fixture(params=[hooks.before, hooks.after])
def register(request):
    return request.param


def test_function_hooks(app: App, client):
    with function_hooks() as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        async def foo(req, res):
            pass

        client.get("/foo")


def test_classed_based_hooks(app: App, client):
    with class_hooks() as (before, after):

        @app.route("/foo")
        class Foo:
            @hooks.before(before)
            @hooks.after(after)
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_class_hook_must_be_callable(app, register):
    class Hook:
        pass

    with pytest.raises(AssertionError) as ctx:
        register(Hook())

    assert "__call__" in str(ctx.value)


def test_async_check(app, register):
    def hook(req, res, params):
        pass

    class Hook:
        def __call__(self, req, res, params):
            pass

    with pytest.raises(ExpectedAsync):
        register(hook)

    with pytest.raises(ExpectedAsync):
        register(Hook)


def test_pass_extra_args(app: App, client):
    with function_hooks(expected_after=1) as (before, after):

        @app.route("/foo")
        class Foo:
            @hooks.before(before, True)  # positional
            @hooks.after(after, value=1)  # keyword
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_use_hook_on_view_class(app: App, client):
    with class_hooks() as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        class Foo:
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_method_hooks(app: App, client):
    with class_hooks() as (before, after):

        @app.route("/foo")
        class Foo:
            @hooks.before(before)
            @hooks.after(after)
            async def get(self, req, res):
                pass

        client.get("/foo")


def test_before_does_not_run_if_method_not_allowed(app: App, client):
    with function_hooks(False, False) as (before, after):

        @app.route("/foo")
        @hooks.before(before)
        @hooks.after(after)
        class Foo:
            async def get(self, req, res):
                pass

        response = client.put("/foo")
        assert response.status_code == 405



import time
from contextlib import contextmanager
from multiprocessing import Value
import typing

import requests


@contextmanager
def function_hooks(
    expected_before: typing.Any = True, expected_after: typing.Any = True
):
    flags = {"before": False, "after": False}

    async def before(req, res, params, value=expected_before):
        nonlocal flags
        assert not flags["after"]
        flags["before"] = value

    async def after(req, res, params, value=expected_after):
        nonlocal flags
        assert flags["before"]
        flags["after"] = value

    yield before, after

    assert flags["before"] is expected_before
    assert flags["after"] is expected_after


@contextmanager
def class_hooks():
    flags = {"before": False, "after": False}

    class SetFlag:
        def __init__(self, flag, value):
            self.flag = flag
            self.value = value

        async def __call__(self, req, res, params):
            nonlocal flags
            flags[self.flag] = self.value

    yield SetFlag("before", True), SetFlag("after", True)

    assert flags["before"]
    assert flags["after"]


def stops_incrementing(
    counter: Value, response: requests.Response, tolerance: int = 15
) -> bool:
    # Check that a counter stops incrementing after the response is closed.

    # tolerance (int):
    # Maximum number of events the server is allowed to send after the
    # connection has been closed.

    def wait_for_events(expect_many=False):
        nonlocal counter
        num_before = counter.value
        time.sleep(0.1)
        num_after = counter.value
        if expect_many:
            assert num_after - num_before >= 5 * tolerance, (
                num_after,
                num_before,
                tolerance,
            )
        return num_after

    sent_before_closing = wait_for_events(expect_many=True)
    response.close()
    sent_after_closing = wait_for_events() - sent_before_closing
    assert sent_after_closing <= tolerance, (sent_after_closing, tolerance)
    return True



import os
import random
from multiprocessing import Event, Process
import typing

import uvicorn
from starlette.testclient import TestClient

try:
    import pytest
except ImportError:  # pragma: no cover
    pytest = None

if typing.TYPE_CHECKING:
    from .applications import App


def create_client(app: "App", **kwargs) -> TestClient:
    """Create a [Starlette Test Client][client] out of an application.

    [client]: https://www.starlette.io/testclient/

    # Parameters
    app: an #::bocadillo.applications#App instance.
    **kwargs (any): keyword arguments passed to the `TestClient` constructor.

    # Returns
    client (TestClient): a Starlette test client.
    """
    return TestClient(app, **kwargs)


class ServerURL(str):
    def __call__(self, path: str) -> str:
        return self + path


class LiveServer:
    """Context manager to spin up a live uvicorn server in a separate process.

    The server process is terminated when exiting the context.

    # Parameters
    app: an #::bocadillo.applications#App instance.
    ready_timeout (float):
        The maximum time to wait for the live server to be ready to handle
        connections, in seconds.
        Defaults to 5.
    stop_timeout (float):
        The maximum time to wait for the live server to terminate, in seconds.
        Defaults to 5.
    **kwargs (any):
        keyword arguments passed to `uvicorn.run()`.
        `host` defaults to `"127.0.0.1."` (i.e. `localhost`).
        `port` defaults to a random integer between 3000 and 9999.
    
    # Attributes
    url (str):
        the full URL where the server lives.
        Build a full URL by calling this attribute,
        e.g. `server.url("/path/foo")`.
    """

    def __init__(
        self,
        app: "App",
        ready_timeout: float = 5,
        stop_timeout: float = 5,
        **kwargs,
    ):
        if (
            pytest is not None and os.getenv("CI") and os.getenv("TRAVIS")
        ):  # pragma: no cover
            pytest.skip("live server process sometimes makes travis ci stall")

        kwargs.setdefault("host", "127.0.0.1")
        kwargs.setdefault("port", random.randint(3000, 9999))

        self.app = app
        self.kwargs = kwargs
        self.ready_timeout = ready_timeout
        self.stop_timeout = stop_timeout
        self._process: Process = None

    @property
    def url(self) -> ServerURL:
        host = self.kwargs["host"]
        port = self.kwargs["port"]
        return ServerURL(f"http://{host}:{port}")

    def __enter__(self):
        ready = Event()

        def target():
            async def callback_notify():
                # Run periodically by the Uvicorn server.
                ready.set()

            uvicorn.run(
                self.app, callback_notify=callback_notify, **self.kwargs
            )

        self._process = Process(target=target)
        self._process.start()

        if not ready.wait(self.ready_timeout):  # pragma: no cover
            raise TimeoutError(
                f"Live server not ready after {self.ready_timeout} seconds"
            )

        return self

    def __exit__(self, *args):
        self._process.terminate()
        self._process.join(self.stop_timeout)
        if self._process.exitcode is None:  # pragma: no cover
            raise TimeoutError(
                f"Live server still running after {self.stop_timeout} seconds."
            )



from contextlib import contextmanager
import typing

from jinja2 import Environment, FileSystemLoader, Template


DEFAULT_TEMPLATES_DIR = "templates"


class Templates:
    """This class provides templating capabilities.

    This is a light wrapper around [jinja2](http://jinja.pocoo.org/docs).

    See also [Templates](../guides/agnostic/templates.md) for detail.

    ::: tip
    Parameters are also stored as attributes and can be accessed or
    modified at runtime.
    :::

    # Parameters
    app (any):
        an optional application object.
        May be a subclass of #::bocadillo.routing#RoutingMixin.
    directory (str):
        the directory where templates should be searched for.
        Defaults to `"templates"` relative to the current working directory.
    context (dict, optional):
        global template variables.
    """

    __slots__ = ("app", "_directory", "_environment")

    def __init__(
        self,
        app: typing.Optional[typing.Any] = None,
        directory: str = DEFAULT_TEMPLATES_DIR,
        context: dict = None,
    ):
        if context is None:
            context = {}

        self.app = app

        self._directory = directory
        self._environment = Environment(
            loader=FileSystemLoader([self.directory]), autoescape=True
        )
        self._environment.globals.update(context)

    @property
    def directory(self) -> str:
        return self._directory

    @directory.setter
    def directory(self, directory: str):
        self._directory = directory
        self._loader.searchpath = [self._directory]  # type: ignore

    @property
    def context(self) -> dict:
        return self._environment.globals

    @context.setter
    def context(self, context: dict):
        self._environment.globals = context

    @property
    def _loader(self) -> FileSystemLoader:
        return typing.cast(FileSystemLoader, self._environment.loader)

    def _get_template(self, name: str) -> Template:
        return self._environment.get_template(name)

    @contextmanager
    def _enable_async(self):
        # Temporarily enable jinja2 async support.
        self._environment.is_async = True
        try:
            yield
        finally:
            self._environment.is_async = False

    async def render(
        self, filename: str, *args: dict, **kwargs: typing.Any
    ) -> str:
        """Render a template asynchronously.

        Can only be used within async functions.

        # Parameters
        name (str):
            name of the template, located inside `templates_dir`.
            The trailing underscore avoids collisions with a potential
            context variable named `name`.
        *args (dict):
            context variables to inject in the template.
        *kwargs (str):
            context variables to inject in the template.
        """
        with self._enable_async():
            return await self._get_template(filename).render_async(
                *args, **kwargs
            )

    def render_sync(
        self, filename: str, *args: dict, **kwargs: typing.Any
    ) -> str:
        """Render a template synchronously.

        # See Also
        [Templates.render](#render) for the accepted arguments.
        """
        return self._get_template(filename).render(*args, **kwargs)

    def render_string(
        self, source: str, *args: dict, **kwargs: typing.Any
    ) -> str:
        """Render a template from a string (synchronously).

        # Parameters
        source (str): a template given as a string.

        # See Also
        [Templates.render](#render) for the other accepted arguments.
        """
        template = self._environment.from_string(source=source)
        return template.render(*args, **kwargs)



from bocadillo import App


def test_mount(app: App, client):
    other = App("other")

    @other.route("/foo")
    async def foo(req, res):
        res.text = "OK"

    app.mount("/other", other)

    r = client.get("/other/foo")
    assert r.status_code == 200
    assert r.text == "OK"



import pytest
import typesystem

from bocadillo import configure, create_client


class Todo(typesystem.Schema):
    title = typesystem.String(max_length=50)
    done = typesystem.Boolean(default=False)


def test_handle_validation_errors(app, client):
    @app.route("/todos")
    class TodoList:
        async def post(self, req, res):
            todo = Todo.validate(await req.json())
            res.json = dict(todo)

    r = client.post("/todos", json={"title": "Make sugar"})
    assert r.status_code == 200
    assert r.json() == {"title": "Make sugar", "done": False}

    r = client.post("/todos", json={"title": "Very long string" * 10})
    assert r.status_code == 400
    assert r.json()["detail"] == {
        "title": "Must have no more than 50 characters."
    }


def test_disable_validation_error_handling(raw_app):
    app = configure(raw_app, handle_typesystem_validation_errors=False)
    client = create_client(app)

    @app.route("/todos")
    class TodoList:
        async def post(self, req, res):
            todo = Todo.validate(await req.json())
            res.json = dict(todo)

    with pytest.raises(typesystem.ValidationError):
        client.post("/todos", json={"title": "Very long string" * 10})



import pytest

from bocadillo import settings, configure, create_client, plugin
from bocadillo.plugins import _PLUGINS


INITIAL_PLUGINS = dict(_PLUGINS)


@pytest.fixture(autouse=True)
def reset_plugins():
    yield
    _PLUGINS.clear()
    for name, value in INITIAL_PLUGINS.items():
        _PLUGINS[name] = value


def test_basic(raw_app):
    @plugin
    def use_foo(app):
        @app.route("/foo")
        async def foo(req, res):
            res.text = "Foo"

    app = configure(raw_app)
    client = create_client(app)

    r = client.get("/foo")
    assert r.status_code == 200
    assert r.text == "Foo"


def test_use_settings(raw_app):
    @plugin
    def use_hello(app):
        hello_message = getattr(settings, "HELLO_MESSAGE")

        @app.route("/hello")
        async def foo(req, res):
            res.text = hello_message

    app = configure(raw_app, hello_message="Hello, plugins!")
    client = create_client(app)

    r = client.get("/hello")
    assert r.status_code == 200
    assert r.text == "Hello, plugins!"



from pathlib import Path

import pytest

from bocadillo import App


@pytest.fixture(name="txt")
def fixture_txt(tmp_path) -> Path:
    txt = tmp_path / "hello.txt"
    txt.write_text("hi files")
    assert txt.name == "hello.txt"
    return txt


@pytest.mark.parametrize("attach", (True, False, None))
def test_file_response(app: App, client, txt: Path, attach: bool):
    kwargs = {} if attach is None else {"attach": attach}

    @app.route("/")
    async def index(req, res):
        res.file(str(txt), **kwargs)

    response = client.get("/")
    assert response.status_code == 200
    assert response.text == txt.read_text()
    if attach is False:
        assert "content-disposition" not in response.headers
    else:
        assert (
            response.headers["content-disposition"]
            == f"attachment; filename='{txt.name}'"
        )


def test_if_file_does_not_exist_then_fail(app: App, client):
    @app.route("/")
    async def index(req, res):
        res.file("doesnotexist.txt")

    with pytest.raises(RuntimeError) as ctx:
        client.get("/")

    assert "does not exist" in str(ctx.value)



import pytest

from bocadillo import configure, create_client


def test_cannot_reconfigure(app):
    with pytest.raises(RuntimeError):
        configure(app)


def test_must_be_configured_to_serve(raw_app):
    @raw_app.route("/")
    async def index(req, res):
        pass

    with pytest.raises(RuntimeError) as ctx:
        with create_client(raw_app):
            pass

    assert "configure(app)" in str(ctx.value)



import typing

import pytest
import typesystem

from bocadillo import create_client, HTTPError
from bocadillo.error_handlers import error_to_json


def setup_http(app, annotation):
    @app.route("/{value}")
    async def index(req, res, value: annotation):
        res.json = {"value": value}


def setup_websocket(app, annotation):
    @app.websocket_route("/{value}")
    async def index(ws, value: annotation):
        await ws.send_json({"value": value})


def get_http_json(client, url, **kwargs) -> dict:
    r = client.get(url, **kwargs)
    assert r.status_code == 200
    return r.json()


def get_websocket_json(client, url, **kwargs) -> dict:
    with client.websocket_connect(url, **kwargs) as ws:
        return ws.receive_json()


@pytest.mark.parametrize("setup, get_json", [(setup_http, get_http_json)])
@pytest.mark.parametrize(
    "annotation, string_value, converted_value",
    [
        (int, "42", 42),
        (int, "4.2", 4),
        (float, "4.2", 4.2),
        (bool, "TRUE", True),
        (bool, "true", True),
        (bool, "1", True),
        (bool, "FALSE", False),
        (bool, "false", False),
        (bool, "0", False),
        (str, "foo", "foo"),
    ],
)
def test_convert_route_parameters(
    app,
    client,
    setup,
    get_json,
    annotation: typing.Type,
    string_value: str,
    converted_value: typing.Any,
):
    setup(app, annotation)
    json = get_json(client, f"/{string_value}")
    assert json["value"] == converted_value


def setup_http_error_handler(app):
    app.add_error_handler(HTTPError, error_to_json)


def setup_websocket_error_handler(app):
    pass


def check_http_status(client, url):
    r = client.get(url)
    assert r.status_code == 400
    return r


def check_websocket_status(client, url):
    with client.websocket_connect(url) as ws:
        message = ws.receive()
        assert message["code"] == 403


@pytest.mark.parametrize(
    "setup, setup_error_handler, check_status",
    [
        (setup_http, setup_http_error_handler, check_http_status),
        (
            setup_websocket,
            setup_websocket_error_handler,
            check_websocket_status,
        ),
    ],
)
@pytest.mark.parametrize(
    "annotation, string_value",
    [(int, "a1"), (float, "foo"), (bool, "12"), (bool, "yes"), (bool, "no")],
)
def test_if_invalid_route_parameter_then_error_response(
    app,
    setup,
    setup_error_handler,
    check_status,
    annotation: typing.Type,
    string_value: str,
):
    setup_error_handler(app)
    setup(app, annotation)
    client = create_client(app, raise_server_exceptions=False)
    r = check_status(client, f"/{string_value}")
    if r is not None:
        assert "value" in r.json()["detail"]


def test_typesystem_converter(app, client):
    @app.route("/{number}")
    async def show(req, res, number: typesystem.Integer(minimum=0)):
        res.json = {"number": number}

    r = client.get("/12")
    assert r.json() == {"number": 12}
    r = client.get("/-12")
    assert r.status_code == 400


def test_defaults_to_str(app, client):
    @app.route("/{number}")
    async def show(req, res, number: "unknown_annotation"):
        res.json = {"number": number}

    r = client.get("/12")
    assert r.json() == {"number": "12"}


def setup_http_query_params_route(app, annotation, default):
    @app.route("/")
    async def index(req, res, value: annotation = default):
        res.json = {"value": value}


def setup_websocket_query_params_route(app, annotation, default):
    @app.websocket_route("/")
    async def index(ws, value: annotation = default):
        await ws.send_json({"value": value})


@pytest.mark.parametrize(
    "setup, get_json",
    [
        (setup_http_query_params_route, get_http_json),
        (setup_websocket_query_params_route, get_websocket_json),
    ],
)
@pytest.mark.parametrize(
    "annotation, default, params, result",
    [
        (int, None, {}, {"value": None}),
        (int, None, {"value": 42}, {"value": 42}),
        (int, 42, {}, {"value": 42}),
        (typesystem.Number(minimum=0), None, {}, {"value": None}),
        (typesystem.Number(minimum=0), 42, {}, {"value": 42}),
    ],
)
def test_querystring_converter(
    app, client, setup, get_json, annotation, default, params, result
):
    setup(app, annotation, default)
    querystring = "?" + "&".join(
        f"{key}={value}" for key, value in params.items()
    )
    json = get_json(client, f"/{querystring}")
    assert json == result


def test_specifiers_fail(app):
    with pytest.raises(TypeError) as ctx:

        @app.route("/{x:d}")
        async def index(req, res, x):
            pass

    # Migration hint must be present.
    error = str(ctx.value)
    assert "pk: int" in error
    assert "{pk:d}" in error



from asyncio import sleep
from contextlib import contextmanager

import pytest

from bocadillo import App, ExpectedAsync, HTTPError, Middleware


def test_async_check(app):
    with pytest.raises(ExpectedAsync):

        class BeforeMiddleware(Middleware):
            def before_dispatch(self):
                pass

    with pytest.raises(ExpectedAsync):

        class AfterMiddleware(Middleware):
            def after_dispatch(self):
                pass


@contextmanager
def build_middleware(
    expect_kwargs=None, expect_call_after=True, old_style=False
):
    called = {"before": False, "after": False}
    kwargs = None

    class SetCalled(Middleware):
        if old_style:

            def __init__(self, inner, app: App, **kw):
                super().__init__(inner, app, **kw)
                nonlocal kwargs
                kwargs = kw
                assert isinstance(app, App)

        else:

            def __init__(self, inner, **kw):
                super().__init__(inner, **kw)
                nonlocal kwargs
                kwargs = kw

        async def before_dispatch(self, req, res):
            nonlocal called
            await sleep(0.01)
            called["before"] = True

        async def after_dispatch(self, req, res):
            nonlocal called
            await sleep(0.01)
            called["after"] = True

    yield SetCalled

    assert called["before"] is True
    assert called["after"] is expect_call_after

    if expect_kwargs is not None:
        kwargs.pop("app", None)  # old-style compatibility
        assert kwargs == expect_kwargs


def test_if_middleware_is_added_then_it_is_called(app: App, client):
    with build_middleware() as middleware:
        app.add_middleware(middleware)

        @app.route("/")
        async def index(req, res):
            pass

        client.get("/")


def test_old_style_middleware(app: App, client):
    with build_middleware(old_style=True) as middleware:
        app.add_middleware(middleware)

        @app.route("/")
        async def index(req, res):
            pass

        client.get("/")


def test_can_pass_extra_kwargs(app: App, client):
    kwargs = {"foo": "bar"}
    with build_middleware(expect_kwargs=kwargs) as middleware:
        app.add_middleware(middleware, **kwargs)

        @app.route("/")
        async def index(req, res):
            pass

        client.get("/")


def test_only_before_dispatch_is_called_if_method_not_allowed(app: App, client):
    with build_middleware(expect_call_after=False) as middleware:
        app.add_middleware(middleware)

        @app.route("/")
        async def index(req, res):
            pass

        response = client.put("/")
        assert response.status_code == 405


@pytest.mark.parametrize("when", ["before", "after"])
def test_errors_raised_in_callback_are_handled(app: App, client, when):
    class CustomError(Exception):
        pass

    @app.error_handler(CustomError)
    async def handle_error(req, res, exception):
        res.text = "gotcha!"

    class MiddlewareWithErrors(Middleware):
        async def before_dispatch(self, req, res):
            if when == "before":
                raise CustomError

        async def after_dispatch(self, req, res):
            if when == "after":
                raise CustomError

    app.add_middleware(MiddlewareWithErrors)

    @app.route("/")
    async def index(req, res):
        pass

    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "gotcha!"


def test_middleware_uses_registered_http_error_handler(app: App, client):
    @app.error_handler(HTTPError)
    async def custom(req, res, exc: HTTPError):
        res.status_code = exc.status_code
        res.text = "Foo"

    class NopeMiddleware(Middleware):
        async def before_dispatch(self, req, res):
            raise HTTPError(401)

    app.add_middleware(NopeMiddleware)

    @app.route("/")
    async def index(req, res):
        pass

    response = client.get("/")
    assert response.status_code == 401
    assert response.text == "Foo"


def test_return_response_in_before_hook(app: App, client):
    class NopeMiddleware(Middleware):
        async def before_dispatch(self, req, res):
            res.text = "Foo"
            return res

    app.add_middleware(NopeMiddleware)

    @app.route("/")
    async def index(req, res):
        # Should not be called
        assert False

    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "Foo"


@pytest.mark.xfail(reason="not supported")
def test_middleware_called_if_routed_to_sub_app(app: App, client):
    with build_middleware() as middleware:
        app.add_middleware(middleware)

        sub = App()

        @sub.route("/home")
        async def home(req, res):
            res.text = "OK"

        app.mount("/sub", sub)

        r = client.get("/sub/home")
        assert r.status_code == 200
        assert r.text == "OK"



from json import dumps
import typing


class server_event(str):
    """A string-like object that represents a [Server-Sent Event][sse].

    [sse]: https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format

    # Examples

    ```python
    >>> from bocadillo import server_event
    # Empty event
    >>> server_event()
    '\\n\\n'
    # Event with data
    >>> server_event("hello, world")
    'data: hello, world\\n\\n'
    # Named event with JSON data
    >>> server_event(name="userconnect", json={"username": "bobby"})
    'event: userconnect\\ndata: {"username": "bobby"}\\n\\n'
    ```

    # Parameters
    name (str):
        An optional `name` for the event.
    data (str or sequence):
        The event `data`. A sequence of strings can be given for
        multi-line event data.
    json (any):
        A JSON-serializable value which, if given, is serialized and used as
        `data`.
    id (int):
        An optional `id` for the event.
    """

    def __new__(
        cls,
        name: str = None,
        *,
        data: typing.Union[str, typing.Sequence] = None,
        json: typing.Any = None,
        id: int = None,
    ):
        if json is not None:
            data = dumps(json)

        parts = [("id", id), ("event", name)]

        if isinstance(data, str):
            parts.append(("data", data))
        elif data is not None:
            for item in data:
                parts.append(("data", item))

        sse_message = (
            "\n".join(
                f"{field}: {value}"
                for field, value in parts
                if value is not None
            )
            + "\n\n"
        )

        return super().__new__(cls, sse_message)

    # For API reference docs and IDE discovery only.
    def __init__(
        self,
        name: str = None,
        *,
        data: typing.Union[str, typing.Sequence] = None,
        json: typing.Any = None,
        id: int = None,
    ):
        pass



import pytest

from bocadillo import App, view
from bocadillo.constants import ALL_HTTP_METHODS


@pytest.mark.parametrize(
    "methods, method, status",
    [(["get", "post"], "get", 200), (["post"], "get", 405), ([], "put", 405)],
)
def test_allowed_methods(app: App, client, methods, method, status):
    @app.route("/")
    @view(methods=methods)
    async def index(req, res):
        pass

    response = getattr(client, method)("/")
    assert response.status_code == status


@pytest.mark.parametrize(
    "method", ["post", "delete", "put", "patch", "options"]
)
def test_unsafe_methods_not_supported_by_default(app: App, client, method):
    @app.route("/")
    async def index(req, res):
        pass

    response = getattr(client, method)("/")
    assert response.status_code == 405


def test_if_get_implemented_then_head_mapped(app: App, client):
    @app.route("/")
    class Index:
        async def get(self, req, res):
            pass

    assert client.head("/").status_code == 200


def test_if_get_in_function_view_methods_then_head_mapped(app: App, client):
    @app.route("/")
    @view(methods=["get"])
    async def index(req, res):
        pass

    assert client.head("/").status_code == 200


def test_if_methods_is_all_then_all_methods_allowed(app: App, client):
    @app.route("/")
    @view(methods=all)
    async def index(req, res):
        pass

    for method in map(str.lower, ALL_HTTP_METHODS):
        assert getattr(client, method)("/").status_code == 200



from http import HTTPStatus
import typing

from .app_types import _E, ErrorHandler, HTTPApp
from .compat import check_async
from .request import Request
from .response import Response


class HTTPError(Exception):
    """Raised when an HTTP error occurs.

    You can raise this within a view or an error handler to interrupt
    request processing.

    # Parameters
    status (int or HTTPStatus):
        the status code of the error.
    detail (any):
        extra detail information about the error. The exact rendering is
        determined by the configured error handler for `HTTPError`.

    # See Also
    - [HTTP response status codes (MDN web docs)](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status)
    """

    __slots__ = ("_status", "detail")

    def __init__(
        self, status: typing.Union[int, HTTPStatus], detail: typing.Any = ""
    ):
        if isinstance(status, int):
            status = HTTPStatus(  # pylint: disable=no-value-for-parameter
                status
            )
        else:
            assert isinstance(
                status, HTTPStatus
            ), f"Expected int or HTTPStatus, got {type(status)}"
        self._status = status
        self.detail = detail

    @property
    def status_code(self) -> int:
        """Return the HTTP error's status code, e.g. `404`."""
        return self._status.value

    @property
    def status_phrase(self) -> str:
        """Return the HTTP error's status phrase, e.g. `"Not Found"`."""
        return self._status.phrase

    @property
    def title(self) -> str:
        """Return the HTTP error's title, e.g. `"404 Not Found"`."""
        return f"{self.status_code} {self.status_phrase}"

    def __str__(self):
        return self.title


class ServerErrorMiddleware(HTTPApp):
    """Return 500 response when an unhandled exception occurs.

    Adaptation of Starlette's `ServerErrorMiddleware`.
    """

    __slots__ = ("app", "handler", "exception")

    def __init__(self, app: HTTPApp, handler: ErrorHandler) -> None:
        self.app = app
        self.handler = handler
        self.exception: typing.Optional[BaseException] = None

    def raise_if_exception(self):
        if self.exception is not None:
            raise self.exception from None

    async def __call__(self, req: Request, res: Response) -> Response:
        try:
            res = await self.app(req, res)
        except BaseException as exc:
            self.exception = exc
            await self.handler(req, res, HTTPError(500))
            return res
        else:
            return res


class HTTPErrorMiddleware(HTTPApp):
    """Handle exceptions that occur while handling HTTP requests.

    Adaptation of Starlette's `ExceptionMiddleware`.
    """

    __slots__ = ("app", "_exception_handlers")

    def __init__(self, app: HTTPApp) -> None:
        self.app = app
        self._exception_handlers: typing.Dict[
            typing.Type[BaseException], ErrorHandler
        ] = {}

    def add_exception_handler(
        self, exception_class: typing.Type[_E], handler: ErrorHandler
    ) -> None:
        assert issubclass(
            exception_class, BaseException
        ), f"expected an exception class, not {type(exception_class)}"
        check_async(
            handler,
            reason=f"error handler '{handler.__name__}' must be asynchronous",
        )
        self._exception_handlers[exception_class] = handler

    def _get_exception_handler(self, exc: _E) -> typing.Optional[ErrorHandler]:
        try:
            return self._exception_handlers[type(exc)]
        except KeyError:
            for cls, handler in self._exception_handlers.items():
                if isinstance(exc, cls):
                    return handler
            return None

    async def __call__(self, req: Request, res: Response) -> Response:
        response = self.app(req, res)

        while True:
            # Deal with errors while there's one.
            # Allows error handlers to raise exceptions to be handled
            # by other error handlers, e.g. raising an `HTTPError` in an
            # error handler.

            has_error = False

            try:
                res = (await response) or res
            except Exception as exc:  # pylint: disable=broad-except
                has_error = True
                handler = self._get_exception_handler(exc)
                if handler is None:
                    raise exc from None
                response = handler(req, res, exc)

            if not has_error:
                assert res is not None
                return res



from bocadillo import configure, create_client


def test_if_hsts_enabled_and_request_is_on_http_then_redirects_to_https(
    raw_app
):
    app = configure(raw_app, hsts=True)

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.get("/", allow_redirects=False)
    assert response.status_code == 301
    assert response.headers["location"] == "https://testserver/"



import pytest

from bocadillo import App


def test_default_response_is_success_empty_text(app: App, client):
    @app.route("/")
    async def index(req, res):
        pass

    response = client.get("/")
    assert response.text == ""
    assert response.status_code == 200


def test_if_status_code_is_no_content_then_no_content_type_set(
    app: App, client
):
    @app.route("/")
    async def index(req, res):
        res.status_code = 204
        pass

    response = client.get("/")
    assert response.status_code == 204
    assert not response.text
    assert response.headers.get("content-type") is None


def test_content_type_defaults_to_plaintext(app: App, client):
    @app.route("/")
    async def index(req, res):
        res.content = "Something magical"
        # make sure no content-type is set before leaving the view
        res.headers.pop("Content-Type", None)
        pass

    response = client.get("/")
    assert response.headers["Content-Type"] == "text/plain"


def test_if_text_set_then_response_is_plain_text(app: App, client):
    @app.route("/")
    async def index(req, res):
        res.text = "foo"
        pass

    response = client.get("/")
    assert response.headers["Content-Type"] == "text/plain"
    assert response.text == "foo"


def test_if_json_set_then_response_is_json(app: App, client):
    @app.route("/")
    async def index(req, res):
        res.json = {"foo": "bar"}

    response = client.get("/")
    assert response.json() == {"foo": "bar"}


def test_if_html_set_then_response_is_html(app: App, client):
    @app.route("/")
    async def index(req, res):
        res.html = "<h1>Foo</h1>"

    response = client.get("/")
    assert response.headers["Content-Type"] == "text/html"
    assert response.text == "<h1>Foo</h1>"


@pytest.mark.parametrize(
    "contents, content_type, check_content",
    [
        [
            [("json", {"foo": "bar"}), ("text", "foo")],
            "text/plain",
            lambda r: r.text == "foo",
        ],
        [
            [("text", "foo"), ("json", {"foo": "bar"})],
            "application/json",
            lambda r: r.json() == {"foo": "bar"},
        ],
    ],
)
def test_last_response_setter_called_has_priority(
    app: App, client, contents: list, content_type: str, check_content
):
    @app.route("/")
    async def index(req, res):
        for attr, value in contents:
            setattr(res, attr, value)

    response = client.get("/")
    assert response.headers["Content-Type"] == content_type
    assert check_content(response)



import decimal
import inspect
from datetime import date, datetime, time
from functools import wraps
import typing

import typesystem

FIELD_ALIASES: typing.Dict[typing.Type, typesystem.Field] = {
    int: typesystem.Integer,
    float: typesystem.Float,
    bool: typesystem.Boolean,
    decimal.Decimal: typesystem.Decimal,
    date: typesystem.Date,
    time: typesystem.Time,
    datetime: typesystem.DateTime,
}


class PathConversionError(typesystem.ValidationError):
    pass


class Converter:

    __slots__ = ("func", "signature", "annotations", "required_params")

    def __init__(self, func: typing.Callable):
        self.func = func
        self.signature = inspect.signature(self.func)

        self.annotations: typing.Dict[str, typing.Type] = {
            param.name: param.annotation
            for param in self.signature.parameters.values()
            if param.annotation is not inspect.Parameter.empty
        }
        self.required_params = set(
            param.name
            for param in self.signature.parameters.values()
            if param.default is inspect.Parameter.empty
        )

    def convert(self, args: tuple, kwargs: dict) -> typing.Tuple[tuple, dict]:
        bound: inspect.BoundArguments = self.signature.bind(*args, **kwargs)

        errors: typing.List[typesystem.ValidationError] = []

        for param_name, value in bound.arguments.items():
            try:
                annotation = self.annotations[param_name]
            except KeyError:
                continue

            # Find the TypeSystem field for the parameter's annotation.
            if isinstance(annotation, typesystem.Field):
                field = annotation
            else:
                try:
                    field = FIELD_ALIASES[annotation]()
                except KeyError:
                    continue

            # Perform validation.
            try:
                value = field.validate(value)
            except typesystem.ValidationError as exc:
                # NOTE: `add_prefix` sets the key of the error in the final
                # error's dict representation.
                errors.extend(exc.messages(add_prefix=param_name))
            else:
                bound.arguments[param_name] = value

        if errors:
            raise PathConversionError(messages=errors)

        # NOTE: apply defaults last to prevent validating the default values.
        # It's faster and less bug-prone.
        bound.apply_defaults()

        return bound.args, bound.kwargs


class ViewConverter(Converter):

    __slots__ = ("query_parameters",)

    def __init__(self, func: typing.Callable):
        super().__init__(func)

        self.query_parameters = set(
            param.name
            for param in self.signature.parameters.values()
            if param.default is not inspect.Parameter.empty
        )

    def get_query_params(self, args: tuple, kwargs: dict) -> dict:
        raise NotImplementedError

    def convert(self, args: tuple, kwargs: dict) -> typing.Tuple[tuple, dict]:
        query_params = self.get_query_params(args, kwargs)

        for param_name in self.query_parameters:
            if param_name in query_params:
                kwargs[param_name] = query_params[param_name]

        return super().convert(args, kwargs)


def convert_arguments(
    func: typing.Callable, converter_class=None
) -> typing.Callable:
    if converter_class is None:
        converter_class = Converter

    converter = converter_class(func)

    @wraps(func)
    async def converted(*args, **kwargs):
        args, kwargs = converter.convert(args, kwargs)
        return await func(*args, **kwargs)

    return converted



import pytest
from bocadillo import App
from jinja2.exceptions import TemplateNotFound

from bocadillo import Templates

from .conftest import TemplateWrapper, create_template


@pytest.mark.asyncio
async def test_render(template_file: TemplateWrapper, templates: Templates):
    html = await templates.render(template_file.name, **template_file.context)
    assert html == template_file.rendered


@pytest.mark.asyncio
async def test_render_using_dict(
    template_file: TemplateWrapper, templates: Templates
):
    html = await templates.render(template_file.name, template_file.context)
    assert html == template_file.rendered


def test_render_sync(template_file: TemplateWrapper, templates: Templates):
    html = templates.render_sync(template_file.name, **template_file.context)
    assert html == template_file.rendered


def test_render_string(templates: Templates):
    html = templates.render_string("<h1>{{ title }}</h1>", title="Hello")
    assert html == "<h1>Hello</h1>"


@pytest.mark.asyncio
async def test_modify_templates_dir(templates: Templates, tmpdir_factory):
    template = create_template(templates, tmpdir_factory, dirname="elsewhere")
    html = await templates.render(template.name, **template.context)
    assert html == template.rendered


@pytest.mark.asyncio
async def test_if_template_does_not_exist_then_not_found_raised(
    templates: Templates
):
    with pytest.raises(TemplateNotFound):
        await templates.render("doesnotexist.html")


def test_use_without_app():
    templates = Templates()
    assert templates.render_string("foo") == "foo"



from bocadillo import App


def test_startup_and_shutdown(app: App, client):
    message = None

    @app.on("startup")
    async def setup():
        nonlocal message
        message = "hi"

    @app.on("shutdown")
    async def cleanup():
        nonlocal message
        message = None

    @app.route("/")
    async def index(req, res):
        res.text = message

    # The Starlette TestClient calls startup and shutdown events when
    # used as a context manager.
    with client:
        assert message == "hi"
        response = client.get("/")
        assert response.text == "hi"
    assert message is None


def test_sync_handler(app: App, client):
    message = None

    @app.on("startup")
    def setup():
        nonlocal message
        message = "hi"

    with client:
        assert message == "hi"


def test_non_decorator_syntax(app: App, client):
    message = None

    async def setup():
        nonlocal message
        message = "hi"

    app.on("startup", setup)

    with client:
        assert message == "hi"



from .applications import App
from .config import settings, SettingsError, configure
from .compat import ExpectedAsync
from .errors import HTTPError
from .middleware import ASGIMiddleware, Middleware
from .injection import discover_providers, provider, useprovider
from .plugins import plugin, get_plugins
from .recipes import Recipe
from .redirection import Redirect
from .request import ClientDisconnect, Request
from .response import Response
from .sse import server_event
from .staticfiles import static
from .templates import Templates
from .testing import create_client, LiveServer
from .views import view
from .websockets import WebSocket, WebSocketDisconnect

__version__ = "0.14.0"



from bocadillo import configure, create_client


def test_if_host_not_allowed_then_400(raw_app):
    app = configure(raw_app, allowed_hosts=["example.com"])
    client = create_client(app)

    @app.route("/")
    async def index(req, res):
        pass

    response = client.get("/")
    assert response.status_code == 400



import inspect
from functools import wraps
import typing

from .compat import check_async
from .request import Request
from .response import Response
from .routing import HTTPRoute
from .views import Handler, get_handlers, View

HookFunction = typing.Callable[
    [Request, Response, dict], typing.Awaitable[None]
]
HookCollection = typing.Dict[HTTPRoute, HookFunction]

BEFORE = "before"
AFTER = "after"


class Hooks:
    """Hooks manager."""

    __slots__ = ()

    def before(self, hook: HookFunction, *args, **kwargs):
        """Register a before hook on a handler.

        # Parameters
        hook (callable): a hook function.
        """
        return self._hook_decorator(BEFORE, hook, *args, **kwargs)

    def after(self, hook: HookFunction, *args, **kwargs):
        """Register an after hook on a handler.

        # Parameters
        hook (callable): a hook function.
        """
        return self._hook_decorator(AFTER, hook, *args, **kwargs)

    @staticmethod
    def _prepare(hook, *args, **kwargs) -> HookFunction:
        class_based = not inspect.isfunction(hook)

        if class_based:
            assert hasattr(
                hook, "__call__"
            ), "class-based hooks must implement __call__()"

        check_target = hook.__call__ if class_based else hook
        name = (
            hook.__class__.__name__ + ".__call__"
            if class_based
            else hook.__name__
        )
        check_async(check_target, reason=f"hook '{name}' must be asynchronous")

        # Enclose args and kwargs
        async def hook_func(req: Request, res: Response, params: dict):
            await hook(req, res, params, *args, **kwargs)

        return hook_func

    def _hook_decorator(
        self, hook_type: str, hook: HookFunction, *args, **kwargs
    ):
        hook = self._prepare(hook, *args, **kwargs)

        def attach_hook(handler: typing.Union[typing.Type[View], Handler]):
            if inspect.isclass(handler):
                # Recursively apply the hook to all handlers.
                for method, _handler in get_handlers(handler).items():
                    setattr(handler, method, attach_hook(_handler))
                return handler
            return _with_hook(hook_type, hook, handler)

        return attach_hook


def _with_hook(hook_type: str, func: HookFunction, handler: Handler):
    async def call_hook(args, kw):
        if len(args) == 2:
            req, res = args
        else:
            # method that has `self` as a first parameter
            req, res = args[1:3]
        assert isinstance(req, Request)
        assert isinstance(res, Response)
        await func(req, res, kw)

    if hook_type == BEFORE:

        @wraps(handler)
        async def with_before_hook(*args, **kwargs):
            await call_hook(args, kwargs)
            await handler(*args, **kwargs)

        return with_before_hook

    assert hook_type == AFTER

    @wraps(handler)
    async def with_after_hook(*args, **kwargs):
        await handler(*args, **kwargs)
        await call_hook(args, kwargs)

    return with_after_hook


# Pre-bind to module
_HOOKS = Hooks()
before = _HOOKS.before
after = _HOOKS.after



import typing

from .request import Request
from .response import Response

if typing.TYPE_CHECKING:
    from .applications import App

_E = typing.TypeVar("_E", bound=Exception)

# ASGI
Scope = dict
Event = typing.MutableMapping[str, typing.Any]
Receive = typing.Callable[[], typing.Awaitable[Event]]
Send = typing.Callable[[Event], None]


class ASGIAppInstance:
    def __call__(self, receive: Receive, send: Send) -> None:
        raise NotImplementedError


class ASGIApp:
    def __call__(self, scope: Scope) -> ASGIAppInstance:
        raise NotImplementedError


# HTTP
Handler = typing.Callable[
    [Request, Response, typing.Any], typing.Awaitable[None]
]
ErrorHandler = typing.Callable[[Request, Response, _E], typing.Awaitable[None]]


class HTTPApp:
    async def __call__(self, req: Request, res: Response) -> Response:
        raise NotImplementedError


# Server lifespan events
EventHandler = typing.Callable[[], typing.Awaitable[None]]



from bocadillo import configure, create_client


def test_no_allowed_origins_by_default(raw_app):
    app = configure(raw_app, cors=True)

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.options(
        "/",
        headers={
            "origin": "foobar.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 400


def test_if_origin_not_in_allow_origins_then_400(raw_app):
    app = configure(raw_app, cors={"allow_origins": ["foobar.com"]})

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.options(
        "/",
        headers={
            "origin": "foobar.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 200

    response = client.options(
        "/",
        headers={
            "origin": "example.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 400


def test_if_method_not_in_allow_methods_then_400(raw_app):
    app = configure(
        raw_app,
        cors={"allow_origins": ["foobar.com"], "allow_methods": ["POST"]},
    )

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.options(
        "/",
        headers={
            "origin": "foobar.com",
            "access-control-request-method": "GET",
        },
    )
    assert response.status_code == 400



import pytest

from bocadillo import App, Recipe, Redirect


def test_if_prefix_not_given_then_routes_mounted_at_slash_name(app, client):
    numbers = Recipe("numbers")

    @numbers.route("/real")
    async def real(req, res):
        pass

    app.recipe(numbers)
    assert client.get("/numbers/real").status_code == 200


def test_if_prefix_then_routes_mounted_at_prefix(app, client):
    numbers = Recipe("numbers", prefix="/numbers-yo")

    @numbers.route("/real")
    async def real(req, res):
        pass

    app.recipe(numbers)
    assert client.get("/numbers-yo/real").status_code == 200


def test_if_prefix_does_not_start_with_slash_then_error_raised():
    with pytest.raises(AssertionError):
        Recipe("numbers", prefix="numbers-yo")


def test_redirect(app: App, client):
    numbers = Recipe("numbers")

    @numbers.route("/R")
    async def R(req, res):
        raise Redirect("/numbers/real")

    @numbers.route("/real")
    async def real(req, res):
        res.text = "inf"

    app.recipe(numbers)

    response = client.get("/numbers/R")
    assert response.status_code == 200
    assert response.text == "inf"



from bocadillo import App


def test_background_task_is_executed(app: App, client):
    executed = False

    @app.route("/")
    async def index(req, res):
        @res.background
        async def notify_executed():
            nonlocal executed
            executed = True

        res.text = "OK"

    response = client.get("/")
    assert response.status_code == 200
    assert response.text == "OK"
    assert executed


def test_background_task_is_executed_after_response_is_sent(app: App, client):
    @app.route("/")
    async def index(req, res):
        @res.background
        async def send_bar():
            res.text = "BAR"

        res.text = "FOO"

    response = client.get("/")
    assert response.text == "FOO"


def test_can_pass_extra_kwargs(app: App, client):
    called = False

    async def set_called(what):
        nonlocal called
        called = what

    @app.route("/")
    async def index(req, res):
        res.background(set_called, "true")

    response = client.get("/")
    assert response.status_code == 200
    assert called == "true"



import pytest

from bocadillo import SettingsError, configure, create_client, view
from bocadillo.utils import override_env


def test_sessions_enabled_no_secret_key(raw_app):
    with pytest.raises(SettingsError):
        configure(raw_app, sessions=True)


@pytest.mark.parametrize("from_env", (True, False))
def test_sessions_enabled_secret_key_empty(raw_app, from_env):
    if from_env:
        with override_env("SECRET_KEY", ""):
            with pytest.raises(SettingsError):
                configure(raw_app, sessions=True)
    else:
        with pytest.raises(SettingsError):
            configure(raw_app, sessions={"secret_key": ""})


@pytest.mark.parametrize("from_env", (True, False))
def test_sessions_enabled_secret_key_present(raw_app, from_env):
    if from_env:
        with override_env("SECRET_KEY", "not-so-secret"):
            app = configure(raw_app, sessions=True)
    else:
        app = configure(raw_app, sessions={"secret_key": "not-so-secret"})

    @app.route("/set")
    @view(methods=["post"])
    async def set_session(req, res):
        req.session["data"] = "something"
        res.text = "Saved"

    @app.route("/")
    async def index(req, res):
        data = req.session["data"]
        res.text = f"Hello {data}"

    client = create_client(app)
    client.post("/set")
    response = client.get("/")
    assert "something" in response.text
    assert "session" in response.cookies



from json import JSONDecodeError
import typing

from starlette.requests import Request as _Request, ClientDisconnect as _CD


ClientDisconnect = _CD


class Request(_Request):
    """The request object, passed to HTTP views and typically named `req`.

    This is a subclass of [`Starlette.requests.Request`][starlette-request]. As a result, all methods and attributes on Starlette's `Request` are available on this class. Additional or redefined members are documented here.

    For usage tips, see [Requests (Guide)](../guides/http/requests.md).

    [starlette-request]: https://www.starlette.io/requests/

    # Methods
    `__aiter__`:
        shortcut for `.stream()`. Allows to process the request body in
        byte chunks using `async for chunk in req: ...`.
    """

    async def json(self) -> typing.Any:
        """Parse the request body as JSON.

        # Returns
        json (dict): the result of `json.loads(await self.body())`.

        # Raises
        HTTPError(400): if the JSON is malformed.
        """
        try:
            return await super().json()
        except JSONDecodeError:
            from .errors import HTTPError  # prevent circular imports

            raise HTTPError(400, detail="JSON is malformed.")

    async def __aiter__(self) -> typing.AsyncGenerator[bytes, None]:
        async for chunk in self.stream():
            yield chunk



from bocadillo import App, Recipe, WebSocket


def test_websocket_recipe_route(app: App, client):
    chat = Recipe("chat")

    @chat.websocket_route("/room/{name}", receive_type="json", send_type="text")
    async def chat_room(ws: WebSocket, name: str):
        message = await ws.receive()
        await ws.send(f"[{name}]: {message['text']}")

    app.recipe(chat)

    with client.websocket_connect("/chat/room/test") as ws:
        ws.send_json({"text": "Hello"})
        assert ws.receive_text() == "[test]: Hello"



from multiprocessing import Value
from time import sleep

import pytest
import requests

from bocadillo import App, ClientDisconnect, LiveServer, server_event

from .utils import stops_incrementing


@pytest.mark.parametrize(
    "args, kwargs, lines",
    [
        [(), {"data": "hello"}, ["data: hello", ""]],
        [(), {"data": "hello", "id": 1}, ["id: 1", "data: hello", ""]],
        [
            ("foo",),
            {"id": 1, "data": "hello"},
            ["id: 1", "event: foo", "data: hello", ""],
        ],
        [
            ("foo",),
            {"id": 1, "json": {"message": "hello"}},
            ["id: 1", "event: foo", 'data: {"message": "hello"}', ""],
        ],
        [(), {"data": ["hello", "world"]}, ["data: hello", "data: world", ""]],
    ],
)
def test_send_event(app: App, client, args: list, kwargs: dict, lines: list):
    @app.route("/events")
    async def view_with_events(req, res):
        @res.event_stream
        async def generate_events():
            yield server_event(*args, **kwargs)

    r = client.get("/events", stream=True)
    assert r.status_code == 200
    stream = r.iter_lines()

    received = [next(stream).decode() for _ in lines]
    assert received == lines

    with pytest.raises(StopIteration):
        next(stream)


def test_sse_headers_are_set(app: App, client):
    @app.route("/events")
    async def view_with_events(req, res):
        @res.event_stream
        async def generate_events():
            yield server_event(name="foo")

    r = client.get("/events")
    assert r.status_code == 200
    assert r.headers["connection"] == "keep-alive"
    assert r.headers["content-type"] == "text/event-stream"
    assert r.headers["cache-control"] == "no-cache"


def test_cache_control_header_not_replaced_if_manually_set(app: App, client):
    @app.route("/events")
    async def sse(req, res):
        res.headers["cache-control"] = "foo"

        @res.event_stream
        async def generate_events():
            yield server_event("hello")

    r = client.get("/events")
    assert r.status_code == 200
    assert r.headers["cache-control"] == "foo"


def test_stop_on_client_disconnect(app: App):
    num_sent = Value("i", 0)

    @app.route("/events")
    async def sse(req, res):
        @res.event_stream
        async def events():
            nonlocal num_sent
            while True:
                yield server_event("hello")
                num_sent.value += 1

    with LiveServer(app) as server:
        r = requests.get(f"{server.url}/events", stream=True)
        assert r.status_code == 200
        assert stops_incrementing(counter=num_sent, response=r)


def test_raise_client_disconnects(app: App):
    caught = Value("i", 0)

    @app.route("/events")
    async def sse(req, res):
        @res.event_stream(raise_on_disconnect=True)
        async def events():
            nonlocal caught
            try:
                while True:
                    yield server_event("hello")
            except ClientDisconnect:
                caught.value = 1

    with LiveServer(app) as server:
        r = requests.get(f"{server.url}/events", stream=True)
        assert r.status_code == 200
        r.close()
        sleep(0.1)
        assert caught.value



from functools import partial
from os.path import basename
import json
import typing

from starlette.background import BackgroundTask
from starlette.requests import Request
from starlette.responses import FileResponse as _FileResponse
from starlette.responses import Response as _Response
from starlette.responses import StreamingResponse as _StreamingResponse

from .constants import CONTENT_TYPE
from .deprecation import deprecated
from .streaming import Stream, StreamFunc, stream_until_disconnect

AnyStr = typing.Union[str, bytes]
BackgroundFunc = typing.Callable[..., typing.Coroutine]


def _content_setter(
    content_type: str, serializer=str, doc: str = None
) -> property:
    def fset(res: "Response", value: AnyStr):
        res.content = serializer(value)
        res.headers["content-type"] = content_type

    if doc is None:
        doc = (
            "Write-only property that sets `content` to the set value and sets "
            f'the `Content-Type` header to `"{content_type}"`.'
        )

    return property(fget=None, fset=fset, doc=doc)


class Response:
    """The response builder, passed to HTTP views and typically named `res`.

    At the lower-level, `Response` behaves like an ASGI application instance:
    it is a callable and accepts `receive` and `send` as defined in the [ASGI
    spec](https://asgi.readthedocs.io/en/latest/specs/main.html#applications).

    [Content-Disposition]: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition

    # Parameters
    request:
        the currently processed #::bocadillo.request#Request.

    # Attributes
    content (bytes or str): the raw response content.
    status_code (int): the HTTP status code. If not set, defaults to `200`.
    headers (dict):
        a case-insensitive dictionary of HTTP headers.
        If not set, `content-type` header is set to `text/plain`.
    chunked (bool): sets the `transfer-encoding` header to `chunked`.
    attachment (str):
        a file name that this response should be sent as.
        This is done by setting the [Content-Disposition] header, and
        typically makes the client browser trigger a "Save As…" dialog or
        download and save the file locally.
    """

    __slots__ = (
        "content",
        "request",
        "status_code",
        "headers",
        "chunked",
        "attachment",
        "_file_path",
        "_background",
        "_stream",
    )

    text = _content_setter(CONTENT_TYPE.PLAIN_TEXT)
    html = _content_setter(CONTENT_TYPE.HTML)
    json = _content_setter(
        CONTENT_TYPE.JSON,
        serializer=json.dumps,
        doc=(
            "Write-only property that sets `content` to the JSON-serialized "
            "version of the set value, "
            'and sets the `Content-Type` header to `"application/json".'
        ),
    )

    def __init__(self, request: Request):
        # Public attributes.
        self.content: typing.Optional[AnyStr] = None
        self.request = request
        self.status_code: typing.Optional[int] = None
        self.headers: typing.Dict[str, str] = {}
        self.chunked = False
        self.attachment: typing.Optional[str] = None
        # Private attributes.
        self._file_path: typing.Optional[str] = None
        self._background: typing.Optional[BackgroundFunc] = None
        self._stream: typing.Optional[Stream] = None

    @deprecated(since="0.14", removal="0.15", alternative="res.json")
    def media(self, value):
        self.json = value

    media = property(fget=None, fset=media)

    def file(self, path: str, attach: bool = True):
        """Send a file asynchronously using [aiofiles].

        [aiofiles]: https://github.com/Tinche/aiofiles

        # Parameters
        path (str):
            a path to a file on this machine.
        attach (bool, optional):
            whether to send the file as an [attachment](#response).
            Defaults to `True`.
        """
        self._file_path = path
        if attach:
            self.attachment = basename(path)

    def background(
        self, func: BackgroundFunc, *args, **kwargs
    ) -> BackgroundFunc:
        """Register a coroutine function to be executed in the background.

        This can be used either as a decorator or a regular function.

        # Parameters
        func (callable):
            a coroutine function.
        *args, **kwargs:
            any positional and keyword arguments to pass to `func` when
            executing it.
        """

        async def background():
            await func(*args, **kwargs)

        self._background = background
        return func

    @property
    def _background_task(self) -> typing.Optional[BackgroundTask]:
        if self._background is not None:
            return BackgroundTask(self._background)
        return None

    def stream(
        self, func: StreamFunc = None, raise_on_disconnect: bool = False
    ) -> StreamFunc:
        """Stream the response.

        The decorated function should be a no-argument asynchronous generator
        function that yields chunks of the response (strings or bytes).

        If `raise_on_disconnect` is `True`,  a `ClientDisconnect` exception
        is raised in the generator when it `yield`s a chunk but the client
        has disconnected. Otherwise, the exception is handled and the stream
        stops.
        """
        if func is None:
            return partial(self.stream, raise_on_disconnect=raise_on_disconnect)

        self._stream = stream_until_disconnect(
            self.request, func(), raise_on_disconnect=raise_on_disconnect
        )

        return func

    def event_stream(self, func: StreamFunc = None, **kwargs) -> StreamFunc:
        """Stream server-sent events.

        The decorated function should be a no-argument asynchronous generator
        function that yields SSE messages (strings or bytes).
        You can use the [server_event](./sse.md#server-event) helper to
        ensure that messages are correctly formatted.

        This is nearly equivalent to `@stream()`. The only difference is that
        this decorator also sets SSE-specific HTTP headers:

        - `Cache-Control: no-cache`
        - `Content-Type: text/event-stream`
        - `Connection: Keep-Alive`

        # See Also
        - [Using server-sent events (MDN)](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events)
        """
        self.headers = {
            "cache-control": "no-cache",
            **self.headers,
            "content-type": "text/event-stream",
            "connection": "keep-alive",
        }
        return self.stream(func, **kwargs)

    async def __call__(self, receive, send):
        """Build and send the response."""
        if self.status_code is None:
            self.status_code = 200

        if self.status_code != 204:
            self.headers.setdefault("content-type", "text/plain")

        if self.chunked:
            self.headers["transfer-encoding"] = "chunked"

        if self.attachment is not None:
            disposition = f"attachment; filename='{self.attachment}'"
            self.headers.setdefault("content-disposition", disposition)

        response_kwargs = {
            "content": self.content,
            "headers": self.headers,
            "status_code": self.status_code,
            "background": self._background_task,
        }

        response_cls = _Response

        if self._file_path is not None:
            response_cls = _FileResponse
            response_kwargs["path"] = self._file_path
            # `FileResponse` will populate the response from `path` and
            # doesn't expect `content` to be passed.
            del response_kwargs["content"]
            # `FileResponse` will set the status code to 200.
            del response_kwargs["status_code"]
        elif self._stream is not None:
            response_cls = _StreamingResponse
            response_kwargs["content"] = self._stream

        response: _Response = response_cls(**response_kwargs)
        await response(receive, send)



import pytest

from bocadillo import App, provider, useprovider, WebSocket


@pytest.fixture(autouse=True)
async def hello_provider():
    @provider
    async def hello_format():
        return "Hello, {who}!"

    @provider
    async def hello(hello_format) -> str:
        return hello_format.format(who="providers")


def test_function_based_view(app: App, client):
    @app.route("/hi")
    async def say_hi(req, res, hello):
        res.text = hello

    r = client.get("/hi")
    assert r.status_code == 200
    assert r.text == "Hello, providers!"


def test_function_based_view_with_route_parameters(app: App, client):
    @app.route("/hi/{who}")
    async def say_hi(req, res, hello_format, who):
        res.text = hello_format.format(who=who)

    r = client.get("/hi/peeps")
    assert r.status_code == 200
    assert r.text == "Hello, peeps!"


def test_class_based_view(app: App, client):
    @app.route("/hi")
    class SayHi:
        async def get(self, req, res, hello):
            res.text = hello

    r = client.get("/hi")
    assert r.status_code == 200
    assert r.text == "Hello, providers!"


def test_class_based_view_with_route_parameters(app: App, client):
    @app.route("/hi/{who}")
    class SayHi:
        async def get(self, req, res, hello_format, who):
            res.text = hello_format.format(who=who)

    r = client.get("/hi/peeps")
    assert r.status_code == 200
    assert r.text == "Hello, peeps!"


@pytest.fixture(autouse=True)
def clients_provider():
    @provider(scope="app")
    async def clients():
        return set()


def test_websocket_clients_example(app: App, client):
    @app.websocket_route("/chat")
    async def chat(ws: WebSocket, clients):
        clients.add(ws)
        await ws.send_text(str(len(clients)))

    @app.route("/clients")
    async def client_count(req, res, clients):
        res.json = {"count": len(clients)}

    with client.websocket_connect("/chat") as ws:
        assert ws.receive_text() == "1"

    r = client.get("/clients")
    assert r.status_code == 200
    assert r.json() == {"count": 1}


@pytest.mark.parametrize("by_name", (False, True))
def test_useprovider(app: App, client, by_name):
    called = False

    @provider
    async def set_called():
        nonlocal called
        called = True

    @app.route("/")
    @useprovider("set_called" if by_name else set_called)
    async def index(req, res):
        pass

    r = client.get("/")
    assert r.status_code == 200
    assert called



"""This module defines classes (both abstract and concrete) that power routing.

**Notations**: this modules uses the following [generic types]:

- `_R` refers to a route object, i.e. an instance of a subclass of [BaseRoute](#baseroute).
- `_V` refers to a view object.

[generic types]: https://docs.python.org/3/library/typing.html#generics
"""

import inspect
import typing

from starlette.websockets import WebSocketClose

from . import views
from .app_types import HTTPApp, Receive, Scope, Send
from .errors import HTTPError
from .redirection import Redirect
from .request import Request
from .response import Response
from .urlparse import Parser
from .views import Handler, HandlerDoesNotExist, View
from .websockets import WebSocket, WebSocketView

# Route generic types.
_R = typing.TypeVar("_R", bound="BaseRoute")  # route
_V = typing.TypeVar("_V")  # view


# Utilities.


class RouteMatch(typing.Generic[_R]):  # pylint: disable=unsubscriptable-object
    """Represents a match between an URL path and a route.

    # Parameters
    route: a route object (subclass of #::bocadillo.routing#BaseRoute).
    params (dict): extracted route parameters.
    """

    __slots__ = ("route", "params")

    def __init__(self, route: _R, params: dict):
        self.route = route
        self.params = params


# Base classes.


class BaseRoute(typing.Generic[_V]):
    """The base route class.

    This is referenced as `_R` in the rest of this module.

    # Parameters
    pattern (str): an URL pattern.
    view (_V):
        a view function or object whose actual type is defined by concrete
        routes.
    """

    __slots__ = ("_pattern", "_parser", "view")

    def __init__(self, pattern: str, view: _V):
        self._parser = Parser(pattern)
        self.view = view

    @property
    def pattern(self) -> str:
        return self._parser.pattern

    def parse(self, path: str) -> typing.Optional[dict]:
        """Parse an URL path against the route's URL pattern.

        # Returns
        params (dict or None):
            If the URL path matches the URL pattern, this is a dictionary
            containing the route parameters, otherwise it is `None`.
        """
        return self._parser.parse(path)

    @classmethod
    def normalize(cls, view: typing.Any) -> _V:
        """Perform any conversion necessary to return a proper view object.

        Not implemented.
        """
        raise NotImplementedError

    @classmethod
    def build(cls, view: _V, pattern: str, **kwargs) -> _R:
        """Create a route out of a normalized view."""
        return cls(view=view, pattern=pattern, **kwargs)

    @classmethod
    def create(cls, view: typing.Any, pattern: str, **kwargs) -> _R:
        """Normalize a view and build and a route instance."""
        view: _V = cls.normalize(view)
        return cls.build(view, pattern=pattern, **kwargs)


class BaseRouter(typing.Generic[_R, _V]):
    """The base router class.

    # Attributes
    routes (dict):
        a mapping of URL patterns to route objects.
    """

    __slots__ = ("routes", "route_class")

    route_class: typing.Type[_R]

    def __init__(self):
        self.routes: typing.Dict[str, _R] = {}

    def _get_key(self, route: _R) -> str:
        # Return the key at which `route` should be stored internally.
        raise NotImplementedError

    def add_route(self, route: _R) -> None:
        """Register a route."""
        self.routes[self._get_key(route)] = route

    def route(self, *args, **kwargs) -> typing.Callable[[typing.Any], _R]:
        """Register a route by decorating a view.

        The decorated function or class will be converted to a proper view using
        [`.normalize()`](#normalize), and then fed to
        [`.add_route()`](#add-route).

        # Parameters
        *args, **kwargs:
            passed to [`.add_route()`](#add-route)
            along with the normalized view.
        """

        def decorate(view: typing.Any) -> _R:
            route = self.route_class.create(view, *args, **kwargs)
            self.add_route(route)

        return decorate

    def match(self, path: str) -> typing.Optional[RouteMatch[_R]]:
        """Attempt to match an URL path against one of the registered routes.

        # Parameters
        path (str): an URL path

        # Returns
        match:
            a #::bocadillo.routing#RouteMatch object if the path matched
            a registered route, `None` otherwise.
        """
        for route in self.routes.values():
            params = route.parse(path)
            if params is not None:
                return RouteMatch(route=route, params=params)
        return None


# HTTP.


class HTTPRoute(BaseRoute[View]):
    """Represents the binding of an URL pattern to an HTTP view.

    Subclass of #::bocadillo.routing#BaseRoute.

    # Parameters
    pattern (str):
        an URL pattern.
    view:
        a #::bocadillo.views#View object.
    name (str):
        the route's name.
    """

    __slots__ = ("name",)

    def __init__(self, pattern: str, view: View, name: str):
        super().__init__(pattern, view)
        self.name = name

    @classmethod
    def normalize(cls, view: typing.Any) -> View:
        """Build a #::bocadillo.views#View object.

        The input, free-form `view` object is converted using the following
        rules:

        - Classes are instanciated (without arguments) and converted with
        [`from_obj()`][obj].
        - Callables are converted with [`from_handler()`][handler].
        - Any other object is interpreted as a view-like object, and converted
        with [`from_obj()`][obj].

        [obj]: ./views.md#from-obj
        [handler]: ./views.md#from-handler

        # Returns
        view:
            a #::bocadillo.views#View object,
            ready to be fed to [`.add_route()`](#add-route).
        """
        if isinstance(view, View):
            return view

        if inspect.isclass(view):
            # View-like class.
            return views.from_obj(view())

        if callable(view):
            # Function-based view.
            # NOTE: here, we ensure backward-compatibility with the routing of
            # function-based views pre-0.9.
            return views.from_handler(view)

        # Treat as a view-like object.
        return views.from_obj(view)

    @classmethod
    def build(
        cls, view: View, pattern: str, name: str = None, **kwargs
    ) -> "HTTPRoute":
        """Build an HTTP route.

        # Parameters
        view:
            a #::bocadillo.views#View object
            obtained via [.normalize()](#normalize-2).
        pattern (str): an URL pattern.
        name (str): a route name (inferred from the view if not given).

        # Returns
        route: an instance of #::bocadillo.routing#HTTPRoute.
        """
        if name is None:
            name = view.name

        return super().build(view=view, pattern=pattern, name=name)

    async def __call__(self, req: Request, res: Response, **params):
        method: str = req.method.lower()

        try:
            handler: Handler = self.view.get_handler(method)
        except HandlerDoesNotExist as e:
            raise HTTPError(405) from e

        await handler(req, res, **params)  # type: ignore


class HTTPRouter(HTTPApp, BaseRouter[HTTPRoute, View]):
    """A router for HTTP routes.

    Subclass of #::bocadillo.routing#BaseRouter.

    Note: routes are stored by `name` instead of `pattern`.
    """

    route_class = HTTPRoute

    def _get_key(self, route: HTTPRoute) -> str:
        # NOTE: this ensures that no two routes stored in this
        # router have the same name.
        return route.name

    async def __call__(self, req: Request, res: Response) -> Response:
        match = self.match(req.url.path)

        if match is None:
            raise HTTPError(status=404)

        try:
            await match.route(req, res, **match.params)
        except Redirect as exc:
            res = exc.response

        return res


# WebSocket.


class WebSocketRoute(BaseRoute[WebSocketView]):
    """Represents the binding of an URL pattern to a WebSocket view.

    Subclass of #::bocadillo.routing#BaseRoute.

    # Parameters
    pattern (str): an URL pattern.
    view (coroutine function):
        Should take as parameters a #::bocadillo.websockets#WebSocket object and
        any extracted route parameters.
    kwargs (any):
        passed when building the #::bocadillo.websockets#WebSocket object.
    """

    __slots__ = ("_ws_kwargs",)

    def __init__(self, pattern: str, view: WebSocketView, **kwargs):
        super().__init__(pattern, view)
        self._ws_kwargs = kwargs

    @classmethod
    def normalize(cls, view: typing.Any) -> WebSocketView:
        return WebSocketView(view)

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send, **params
    ):
        ws = WebSocket(scope, receive=receive, send=send, **self._ws_kwargs)
        await self.view(ws, **params)


class WebSocketRouter(BaseRouter[WebSocketRoute, WebSocketView]):
    """A router for WebSocket routes.

    Subclass of #::bocadillo.routing#BaseRouter.
    """

    route_class = WebSocketRoute

    def _get_key(self, route: WebSocketRoute) -> str:
        return route.pattern

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        # Dispatch a WebSocket connection request.
        match = self.match(scope["path"])
        if match is None:
            # Close with a 403 error code, as specified in the ASGI spec:
            # https://asgi.readthedocs.io/en/latest/specs/www.html#close
            await WebSocketClose(code=403)(receive, send)
        else:
            await match.route(scope, receive, send, **match.params)


class RoutingMixin:
    """Provide HTTP and WebSocket routing to an application class."""

    __slots__ = ("http_router", "websocket_router")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.http_router = HTTPRouter()
        self.websocket_router = WebSocketRouter()

    def route(self, pattern: str, *, name: str = None):
        """Register an HTTP route by decorating a view.

        # Parameters
        pattern (str): an URL pattern.
        name (str):
            an optional name for the route.
            If a route already exists for this name, it is replaced.
            Defaults to a snake-cased version of the view's name.
        """
        return self.http_router.route(pattern=pattern, name=name)

    def websocket_route(
        self,
        pattern: str,
        *,
        auto_accept: bool = True,
        value_type: str = None,
        receive_type: str = None,
        send_type: str = None,
        caught_close_codes: typing.Tuple[int] = None,
    ):
        """Register a WebSocket route by decorating a view.

        # Parameters
        pattern (str): an URL pattern.

        # See Also
        - #::bocadillo.websockets#WebSocket for a description of keyword
        arguments.
        """
        # NOTE: use named keyword arguments instead of `**kwargs` to improve
        # their accessibility (e.g. for IDE discovery).
        return self.websocket_router.route(
            pattern=pattern,
            auto_accept=auto_accept,
            value_type=value_type,
            receive_type=receive_type,
            send_type=send_type,
            caught_close_codes=caught_close_codes,
        )



from bocadillo import App


def test_chunked_response(app: App, client):
    @app.route("/")
    async def index(req, res):
        res.chunked = True

    r = client.get("/")
    assert r.headers["transfer-encoding"] == "chunked"



import os
import typing

import typesystem
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware

from .config import SettingsError, settings
from .constants import DEFAULT_CORS_CONFIG
from .converters import PathConversionError
from .error_handlers import error_to_json
from .errors import HTTPError
from .staticfiles import static
from .injection import STORE

if typing.TYPE_CHECKING:
    from .applications import App

PluginFunction = typing.Callable[["App"], None]

_PLUGINS: typing.Dict[str, PluginFunction] = {}


def plugin(func: PluginFunction):
    """Register a new plugin."""
    _PLUGINS[func.__name__] = func
    return func


def get_plugins() -> typing.Dict[str, PluginFunction]:
    """Return the currently registered plugins."""
    return _PLUGINS


@plugin
def use_providers(app: "App"):
    """Configure providers.

    This plugin is always enabled. It ensures that app-scoped providers
    are correctly setup on app startup, and tore down on app shutdown, and
    resolves dependencies between providers.
    """
    STORE.freeze()
    app.on("startup", STORE.enter_session)
    app.on("shutdown", STORE.exit_session)


@plugin
def use_allowed_hosts(app: "App"):
    """Restrict which hosts an application is allowed to be served at.

    Settings:
    - `ALLOWED_HOSTS` (list of str, optional):
        a list of hosts. If the list contains `"*"`, any host is allowed.
        Defaults to `["*"]`.
    """
    allowed_hosts = settings.get("ALLOWED_HOSTS")
    if allowed_hosts is None:
        allowed_hosts = ["*"]

    app.add_asgi_middleware(TrustedHostMiddleware, allowed_hosts=allowed_hosts)


@plugin
def use_cors(app: "App"):
    """Enable CORS (Cross-Origin Resource Sharing) headers.

    [constants.py]: /api/constants.md
    [CORSMiddleware]: https://www.starlette.io/middleware/#corsmiddleware

    Settings:
    - `CORS` (bool or dict):
        if `True`, the default configuration defined in [constants.py] is used.
        Otherwise, the dictionary is passed to Starlette's [CORSMiddleware].
    """
    cors: typing.Optional[typing.Union[bool, dict]] = settings.get("CORS")

    if cors is None:
        return

    if cors is True:
        cors = dict(DEFAULT_CORS_CONFIG)

    app.add_asgi_middleware(CORSMiddleware, **cors)


@plugin
def use_gzip(app: "App"):
    """Enable [GZip] compression.

    [GZip]: /guides/http/middleware.md#gzip

    Settings:
    - `GZIP` (bool):
        if `True`, automatically compress responses for clients that support it.
        Defaults to `False`.
    - `GZIP_MIN_SIZE` (int):
        compress only responses that have more bytes than the specified value.
        Defaults to `1024`.
    """
    if not settings.get("GZIP", False):
        return

    gzip_min_size = settings.get("GZIP_MIN_SIZE", 1024)
    app.add_asgi_middleware(GZipMiddleware, minimum_size=gzip_min_size)


@plugin
def use_hsts(app: "App"):
    """Enable [HSTS].

    [HSTS]: /guides/http/middleware.md#hsts

    Settings:
    - `HSTS` (bool):
        if `True`, HTTP traffic is automatically redirected to HTTPS.
        Defaults to `False`.
    """
    if not settings.get("HSTS"):
        return

    app.add_asgi_middleware(HTTPSRedirectMiddleware)


@plugin
def use_sessions(app: "App"):
    """Enable cookie-based signed sessions.

    [SessionMiddleware]: https://www.starlette.io/middleware/#sessionmiddleware

    Settings:
    - `SESSIONS` (bool or dict):
        if `True`, the secret key is obtained from the `SECRET_KEY` environment
        variable. Otherwise, it must be a dictionary which will be passed
        to Starlette's [SessionMiddleware].
    """
    sessions = settings.get("SESSIONS")

    if sessions is None:
        return

    try:
        from starlette.middleware.sessions import SessionMiddleware
    except ImportError as exc:  # pragma: no cover
        if "itsdangerous" in str(exc):
            raise ImportError(
                "Please install the [sessions] extra to use sessions: "
                "`pip install bocadillo[sessions]`."
            ) from exc
        raise exc from None

    if sessions is True:
        sessions = {"secret_key": os.getenv("SECRET_KEY")}

    if not sessions.get("secret_key"):
        raise SettingsError(
            "`SESSIONS` must have a non-empty `secret_key` to use sessions."
        )

    app.add_asgi_middleware(SessionMiddleware, **sessions)


@plugin
def use_staticfiles(app: "App"):
    """Enable static files serving with WhiteNoise.

    Settings:
    - `STATIC_DIR` (str):
        the name of the directory containing static files, relative to
        the application entry point.
        Set to `None` to not serve any static files.
        Defaults to `"static"`.
    - `STATIC_ROOT` (str):
        the path prefix for static assets. Defaults to `"static"`.
    - `STATIC_CONFIG` (dict):
        extra static files configuration attributes.
        See also #::bocadillo.staticfiles#static.
    """
    static_root = settings.get("STATIC_ROOT", "static")
    static_dir = settings.get("STATIC_DIR", "static")
    static_config = settings.get("STATIC_CONFIG", {})

    if static_dir is None:
        return

    app.mount(static_root, static(static_dir, **static_config))


@plugin
def use_path_conversion_error_handling(app: "App"):
    @app.error_handler(PathConversionError)
    async def on_path_conversion_error(req, res, exc: PathConversionError):
        raise HTTPError(400, detail=dict(exc))


@plugin
def use_typesystem_validation_error_handling(app: "App"):
    """Setup an error handler for `typesystem.ValidationError`.

    This plugin allows to validate TypeSystem schemas while letting the
    framework deal with formatting and sending back a `400 Bad Request`
    error response in case of invalid data.

    Settings:
    - `HANDLE_TYPESYSTEM_VALIDATION_ERRORS` (bool):
        Set to `False` to disable this plugin. Defaults to `True`.
    """
    if not settings.get("HANDLE_TYPESYSTEM_VALIDATION_ERRORS", True):
        return

    @app.error_handler(typesystem.ValidationError)
    async def handle_validation_error(req, res, exc):
        detail = dict(exc)
        # Don't just re-raise `HTTPError` for handling by the configured
        # error handler, because we really want to return JSON.
        await error_to_json(req, res, HTTPError(400, detail=detail))



import traceback
import typing

import typesystem
from starlette.datastructures import URL, Headers, QueryParams
from starlette.websockets import WebSocket as StarletteWebSocket
from starlette.websockets import WebSocketDisconnect as _WebSocketDisconnect

from .app_types import Event, Receive, Scope, Send
from .compat import asyncnullcontext, check_async
from .constants import WEBSOCKET_CLOSE_CODES
from .converters import ViewConverter, convert_arguments
from .injection import consumer


class WebSocket:
    """Represents a WebSocket connection.

    See also [WebSockets](/guides/websockets/).

    # Parameters
    scope (dict): ASGI scope.
    receive (callable): ASGI receive function.
    send (callable): ASGI send function.
    auto_accept (bool):
        whether to automatically accept the WebSocket connection request.
        Defaults to `True`. If `False` is passed, the connection must be 
        accepted via `.accept()` or an `async with` block. Useful to
        perform conditionally `.reject()` a connection request or perform
        advanced error handling.
    value_type (str):
        The type of messages received or sent over the WebSocket.
        If given, overrides `receive_type` and `send_type`.
        Defaults to `None`.
    receive_type (str):
        The type of messages received over the WebSocket.
        Defaults to `"text"`.
    send_type (str):
        The type of messages send over the WebSocket.
        Defaults to `"text"`.
    caught_close_codes (tuple of int):
        Close codes of `WebSocketDisconnect` exceptions that should be
        caught and silenced. Defaults to `(1000, 1001)`.

    # Attributes
    url (str-like): the URL which the WebSocket connection was made to.
    headers (dict): headers attached to the connection request.
    query_params (dict): parsed query parameters.
    """

    __slots__ = (
        "_ws",
        "caught_close_codes",
        "auto_accept",
        "__accepted",
        "receive_type",
        "send_type",
    )

    __default_receive_type__ = "text"
    __default_send_type__ = "text"

    url: URL
    headers: Headers
    query_params: QueryParams

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        auto_accept: bool = True,
        value_type: str = None,
        receive_type: str = None,
        send_type: str = None,
        caught_close_codes: typing.Tuple[int, ...] = None,
    ):
        # NOTE: we use composition over inheritance here, because
        # we want to redefine `receive()` and `send()` but Starlette's
        # WebSocket class uses those in many other functions, which we
        # do not need / want to re-implement.
        # The compromise is the definition of delegated methods below.
        self._ws = StarletteWebSocket(scope, receive=receive, send=send)

        if caught_close_codes is None:
            caught_close_codes = (1000, 1001)
        if caught_close_codes is all:
            caught_close_codes = tuple(WEBSOCKET_CLOSE_CODES)
        self.caught_close_codes = caught_close_codes

        self.auto_accept = auto_accept
        self.__accepted = False

        if value_type is not None:
            receive_type = send_type = value_type
        else:
            receive_type = receive_type or self.__default_receive_type__
            send_type = send_type or self.__default_send_type__

        self.receive_type = receive_type
        self.send_type = send_type

    def __getattr__(self, name: str) -> typing.Any:
        return getattr(self._ws, name)

    def __getitem__(self, name: str) -> typing.Any:
        return self._ws.__getitem__(name)

    # Connection handling.

    async def accept(self, subprotocol: str = None) -> None:
        """Accept the connection request.

        # Parameters
        subprotocol (str): a specific WebSocket subprotocol that should be expected from clients. Defaults to `None`.
        """
        return await self._ws.accept(subprotocol=subprotocol)

    async def reject(self):
        """Reject the connection request.

        This is equivalent to `.close(403)`.
        Calling this before `.accept()` has undefined behavior.
        """
        return await self.close(code=403)

    async def close(self, code: int = 1000) -> None:
        """Close the WebSocket connection.

        # Parameters
        code (int): a close code, defaults to `1000` (Normal Closure).

        # See Also
        - [CloseEvent specification](https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent)
        """
        return await self._ws.close(code=code)

    async def ensure_closed(self, code: int = 1000):
        """Close the connection if it has not been closed already."""
        try:
            await self.close(code=code)
        except RuntimeError:
            # Already closed.
            pass

    # Data reception/sending.

    async def receive_text(self) -> str:
        """Receive a message as text."""
        return await self._ws.receive_text()

    async def send_text(self, data: str):
        """Send a message as text."""
        return await self._ws.send_text(data)

    async def receive_bytes(self) -> bytes:
        """Receive a message a bytes."""
        return await self._ws.receive_bytes()

    async def send_bytes(self, data: bytes):
        """Send a message as bytes."""
        return await self._ws.send_bytes(data)

    async def receive_json(self) -> typing.Union[dict, list]:
        """Receive a message as text and parse it to a JSON object.

        # Raises
        json.JSONDecodeError: if the received JSON is invalid.
        """
        return await self._ws.receive_json()

    async def send_json(self, data: typing.Union[dict, list]):
        """Serialize an object to JSON and send it as text.

        # Raises
        TypeError: if the given `data` is not JSON serializable.
        """
        return await self._ws.send_json(data)

    async def receive_event(self) -> Event:
        """Receive a raw ASGI event."""
        return await self._ws.receive()

    async def send_event(self, event: Event):
        """Send a raw ASGI event."""
        return await self._ws.send(event)

    async def receive(self) -> typing.Union[str, bytes, list, dict]:
        """Receive a message from the WebSocket.

        Shortcut for `receive_<self.receive_type>`.
        """
        receiver = getattr(self, f"receive_{self.receive_type}")
        return await receiver()

    async def send(self, message: typing.Any):
        """Send a message over the WebSocket.

        Shortcut for `send_<self.send_type>`.
        """
        sender = getattr(self, f"send_{self.send_type}")
        return await sender(message)

    # Asynchronous context manager.

    async def __aenter__(self, *args, **kwargs):
        if not self.__accepted:
            await self.accept()
            self.__accepted = True
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type == WebSocketDisconnect:
            # Client has closed the connection.
            # Returning `True` here silences the exception. See:
            # https://docs.python.org/3/reference/datamodel.html#object.__exit__
            return exc_val.code in self.caught_close_codes

        # Close with Internal Error if an unknown exception was raised.
        code = 1000 if exc_type is None else 1011
        # NOTE: view may have closed the WebSocket already, so we
        # must use ensure_closed().
        await self.ensure_closed(code)

    # Asynchronous iterator.

    async def __aiter__(self):
        while True:
            yield await self.receive()


class WebSocketConverter(ViewConverter):
    def get_query_params(self, args: tuple, kwargs: dict) -> dict:
        ws: WebSocket = args[0]
        return ws.query_params


class WebSocketView:

    __slots__ = ("func",)

    def __init__(self, func: typing.Callable):
        check_async(
            func,
            reason=f"WebSocket view '{func.__name__}' must be asynchronous",
        )
        func = convert_arguments(func, converter_class=WebSocketConverter)
        func = consumer(func)
        self.func = func

    async def __call__(self, ws: WebSocket, **params):
        context = ws if ws.auto_accept else asyncnullcontext()
        try:
            async with context:
                try:
                    await self.func(ws, **params)  # type: ignore
                except typesystem.ValidationError:
                    await ws.ensure_closed(403)
                    traceback.print_exc()
        except BaseException:
            traceback.print_exc()
            await ws.ensure_closed(1011)
            raise


WebSocketDisconnect = _WebSocketDisconnect



import os
from contextlib import contextmanager


@contextmanager
def override_env(variable: str, value: str):
    """Context manager to temporarily override an environment variable.

    The variable is restored to its initial value (if any) when exiting the
    context.

    # Example

    ```python
    from bocadillo.utils import override_env

    with override_env("SOME_VAR", "yes"):
        ...
    ```

    # Parameters
    variable (str): the name of the environment variable.
    value (str): a temporary value for the environment variable.
    """
    initial = os.environ.get(variable, None)
    os.environ[variable] = value
    try:
        yield
    finally:
        os.environ.pop(variable)
        if initial is not None:
            os.environ[variable] = initial



import typing

from .app_types import ASGIApp, ASGIAppInstance, HTTPApp, Scope
from .compat import check_async
from .request import Request
from .response import Response

if typing.TYPE_CHECKING:  # pragma: no cover
    from .applications import App


class MiddlewareMeta(type):
    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        for callback in "before_dispatch", "after_dispatch":
            check_async(
                getattr(cls, callback),
                reason="'{name}.{callback}' must be asynchronous",
            )
        return cls


class Middleware(HTTPApp, metaclass=MiddlewareMeta):
    """Base class for middleware classes.

    # Parameters
    inner (callable): the inner middleware that this middleware wraps.
    app: the #::bocadillo.applications#App instance.
    kwargs (any):
        Keyword arguments passed when registering the middleware on `app`.
    """

    def __init__(self, inner: HTTPApp, app: "App" = None, **kwargs):
        # NOTE: app defaults to `None` to support old-style HTTP middleware.
        self.inner = inner
        self.app = app
        self.kwargs = kwargs

    async def before_dispatch(
        self, req: Request, res: Response
    ) -> typing.Optional[Response]:
        """Perform processing before a request is dispatched.

        If the #::bocadillo.response#Response object is returned,
        it will be used and no further processing will be performed.

        # Parameters
        req: a #::bocadillo.request#Request object.
        res: a #::bocadillo.response#Response object.

        # Returns
        res: an optional #::bocadillo.response#Response object.
        """

    async def after_dispatch(self, req: Request, res: Response) -> None:
        """Perform processing after a request has been dispatched.

        # Parameters
        req: a #::bocadillo.request#Request object.
        res: a #::bocadillo.response#Response object.
        """

    async def __call__(self, req: Request, res: Response) -> Response:
        before_res = await self.before_dispatch(req, res)

        if before_res is not None:
            return before_res

        res = await self.inner(req, res)

        await self.after_dispatch(req, res)

        return res


class ASGIMiddleware(ASGIApp):
    """Base class for ASGI middleware classes.

    # Parameters
    inner (callable): the inner middleware.
    app (App): the application instance.
    kwargs (any):
        Keyword arguments passed when registering the middleware on `app`.
    """

    def __init__(self, inner: ASGIApp, app: "App", **kwargs):
        self.inner = inner
        self.app = app
        self.kwargs = kwargs

    def __call__(self, scope: Scope) -> ASGIAppInstance:
        return self.inner(scope)



import inspect
import re
import typing

try:
    # >= 3.7
    from contextlib import nullcontext  # pylint: disable=unused-import
except ImportError:  # pragma: no cover
    from contextlib import contextmanager

    @contextmanager
    def nullcontext(enter_result: typing.Any = None):
        yield enter_result


_CAMEL_REGEX = re.compile(r"(.)([A-Z][a-z]+)")
_SNAKE_REGEX = re.compile(r"([a-z0-9])([A-Z])")

_V = typing.TypeVar("_V")


class ExpectedAsync(Exception):
    """Raised when a feature expected an asynchronous function."""


def check_async(func, reason: str):
    if not inspect.iscoroutinefunction(func):
        message = f"{reason}. Hint: use `async def` instead of `def`."
        raise ExpectedAsync(message)


# For 3.6 compatibility (only available in 3.7+).
class asyncnullcontext:
    def __init__(self, enter_result: typing.Any = None):
        self._result = enter_result

    async def __aenter__(self):
        return self._result

    async def __aexit__(self, *args):
        pass


def camel_to_snake(name: str) -> str:
    """Convert a `CamelCase` name to its `snake_case` version."""
    s1 = _CAMEL_REGEX.sub(r"\1_\2", name)
    return _SNAKE_REGEX.sub(r"\1_\2", s1).lower()


# WSGI

Environ = dict
StartResponse = typing.Callable[[str, typing.List[str]], None]
WSGIApp = typing.Callable[[Environ, StartResponse], typing.List[bytes]]


def empty_wsgi_app() -> WSGIApp:
    """Return a WSGI app that always returns `404 Not Found`."""

    def wsgi(environ, start_response):
        status = "404 Not Found"
        body = b"Not Found"
        headers = [("Content-Type", "text/plain")]
        start_response(status, headers)
        return [body]

    return wsgi



import pytest
from bocadillo import App


def test_json_response(app: App, client):
    data = {"message": "hello"}

    @app.route("/")
    async def index(req, res):
        res.json = data

    response = client.get("/")
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json"
    assert response.json() == data


def test_deprecated_media(app, client):
    data = {"message": "hello"}

    @app.route("/")
    async def index(req, res):
        with pytest.deprecated_call():
            res.media = data

    response = client.get("/")
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json"
    assert response.json() == data



import re
import typing

PARAM_RE = re.compile(r"{}|{([a-zA-Z_:][a-zA-Z0-9_:]*)}")
WILDCARD = "{}"


def compile_path(pattern: str) -> typing.Tuple[typing.Pattern, str]:
    regex = "^"
    path_format = ""
    idx = 0

    for match in PARAM_RE.finditer(pattern):
        name, = match.groups(default="")
        if ":" in name:
            raise TypeError(
                "path specifiers were removed in 0.14.0. "
                "Use type annotations on view parameters instead, "
                "e.g. `pk: int` instead of '{pk:d}'"
            )

        regex += pattern[idx : match.start()]
        expr = r".+" if not name else rf"?P<{name}>[^/]+"
        regex += rf"({expr})"

        path_format += pattern[idx : match.start()]
        path_format += "{%s}" % name

        idx = match.end()

    regex += pattern[idx:] + "$"
    path_format += pattern[idx:]

    return re.compile(regex), path_format


class Parser:
    def __init__(self, pattern: str):
        if pattern != WILDCARD and not pattern.startswith("/"):
            pattern = f"/{pattern}"
        self.regex, self.pattern = compile_path(pattern)

    def parse(self, value: str) -> typing.Optional[dict]:
        match = self.regex.match(value)
        if match is None:
            return None
        return match.groupdict()



"""Meta classes."""


class DocsMeta(type):
    """Metaclass that provides a few fixes related to Pydoc-Markdown."""

    def __new__(mcs, name, bases, namespace):
        mcs._prepare_for_docs(bases, namespace)
        cls = super().__new__(mcs, name, bases, namespace)
        return cls

    @classmethod
    def _prepare_for_docs(mcs, bases, namespace):
        # Pydoc-Markdown only takes documented members from the class'
        # `__dict__`. This is not added automatically when subclassing, so
        # we need to force it.
        for base in bases:
            for key, value in base.__dict__.items():
                if key.startswith("_"):
                    # Don't include private attributes and methods.
                    continue
                if key in namespace:
                    # Don't prevent overriding on application class.
                    continue
                namespace[key] = value



from aiodine import Store, scopes

# pylint: disable=invalid-name
STORE = Store(
    scope_aliases={"request": scopes.FUNCTION, "app": scopes.SESSION},
    providers_module="providerconf",
    default_scope=scopes.FUNCTION,
)
provider = STORE.provider
discover_providers = STORE.discover
useprovider = STORE.useprovider
consumer = STORE.consumer



import pytest

from bocadillo import App, ExpectedAsync, view
from bocadillo.constants import ALL_HTTP_METHODS


def test_function_based_view(app: App, client):
    @app.route("/")
    @view()
    async def index(req, res):
        pass

    assert client.get("/").status_code == 200


def test_async_check(app):
    def index(req, res):
        pass

    with pytest.raises(ExpectedAsync):
        app.route("/")(index)

    class Index:
        def get(self, req, res):
            pass

    with pytest.raises(ExpectedAsync):
        app.route("/")(Index)


def test_view_decorator_is_optional(app: App, client):
    @app.route("/")
    async def index(req, res):
        pass

    assert client.get("/").status_code == 200


def test_can_register_class_based_view(app: App):
    @app.route("/")
    class Index:
        pass


@pytest.mark.parametrize("method", map(str.lower, ALL_HTTP_METHODS))
def test_if_method_not_implemented_then_405(app: App, client, method: str):
    @app.route("/")
    class Index:
        pass

    response = getattr(client, method)("/")
    assert response.status_code == 405


def test_if_method_implemented_then_as_normal(app: App, client):
    @app.route("/")
    class Index:
        async def get(self, req, res):
            res.text = "Get!"

    response = client.get("/")
    assert response.status_code == 200
    assert response.text == "Get!"


def test_if_handle_is_implemented_then_bypasses_other_methods(app: App, client):
    @app.route("/")
    class Index:
        async def handle(self, req, res):
            res.text = "Handle!"

        async def get(self, req, res):
            res.text = "Get!"

    response = client.get("/")
    assert response.status_code == 200
    assert response.text == "Handle!"


def test_view_from_obj(app: App, client):
    class MyView:
        async def get(self, req, res):
            pass

    app.route("/")(MyView())
    assert client.get("/").status_code == 200


def test_parameter_is_passed_as_keyword_argument(app: App, client):
    @app.route("/greet/{person}")
    async def greet(req, res, *, person: str):
        res.text = person

    response = client.get("/greet/John")
    assert response.status_code == 200
    assert response.text == "John"



from bocadillo import configure, create_client


def test_if_gzip_enabled_then_response_is_compressed(raw_app):
    app = configure(raw_app, gzip=True, gzip_min_size=0)

    @app.route("/")
    async def index(req, res):
        pass

    client = create_client(app)
    response = client.get("/", headers={"Accept-Encoding": "gzip"})
    assert response.status_code == 200
    assert response.headers["content-encoding"] == "gzip"



from bocadillo import App


def test_index_returns_404_by_default(app: App, client):
    assert client.get("/").status_code == 404


def test_if_route_not_registered_then_404(app: App, client):
    assert client.get("/test").status_code == 404


def test_if_route_registered_then_not_404(app: App, client):
    @app.route("/")
    async def index(req, res):
        pass

    assert client.get("/").status_code != 404


def test_default_status_code_is_200_on_routes(app: App, client):
    @app.route("/")
    async def index(req, res):
        pass

    assert client.get("/").status_code == 200


def test_leading_slash_is_added_if_not_present(app: App, client):
    @app.route("foo")
    async def index(req, res):
        pass

    assert client.get("/foo").status_code == 200



import typing

from starlette.middleware.wsgi import WSGIResponder
from starlette.routing import Lifespan

from .app_types import (
    _E,
    ASGIApp,
    ASGIAppInstance,
    ErrorHandler,
    EventHandler,
    Receive,
    Scope,
    Send,
)
from .config import settings
from .compat import WSGIApp
from .error_handlers import error_to_text
from .errors import HTTPError, HTTPErrorMiddleware, ServerErrorMiddleware
from .injection import STORE
from .meta import DocsMeta
from .middleware import ASGIMiddleware
from .request import Request
from .response import Response
from .routing import RoutingMixin

if typing.TYPE_CHECKING:  # pragma: no cover
    from .recipes import Recipe


class App(RoutingMixin, metaclass=DocsMeta):
    """The all-mighty application class.

    This class implements the [ASGI](https://asgi.readthedocs.io) protocol.

    # Example

    ```python
    >>> from bocadillo import App
    >>> app = App()
    ```

    # Parameters
    name (str):
        An optional name for the app.
    """

    __slots__ = (
        "name",
        "asgi",
        "exception_middleware",
        "server_error_middleware",
        "_children",
        "_lifespan",
    )

    def __init__(self, name: str = None):
        super().__init__()

        self.name = name

        # Base ASGI app
        self.asgi = self.dispatch

        # Mounted (children) apps.
        self._children: typing.Dict[str, typing.Any] = {}

        # HTTP middleware
        self.exception_middleware = HTTPErrorMiddleware(self.http_router)
        self.server_error_middleware = ServerErrorMiddleware(
            self.exception_middleware, handler=error_to_text
        )
        self.add_error_handler(HTTPError, error_to_text)

        # Lifespan middleware
        self._lifespan = Lifespan()

        # Startup checks.
        @self.on("startup")
        async def check_app():
            if not settings.configured:
                raise RuntimeError(
                    "You must call `configure(app)` before serving `app`. "
                )

        # NOTE: discover providers from `providerconf` at instanciation time,
        # so that further declared views correctly resolve providers.
        STORE.discover_default()

    def mount(self, prefix: str, app: typing.Union["App", ASGIApp, WSGIApp]):
        """Mount another WSGI or ASGI app at the given prefix.

        [WSGI]: https://wsgi.readthedocs.io
        [ASGI]: https://asgi.readthedocs.io

        # Parameters
        prefix (str):
            A path prefix where the app should be mounted, e.g. `"/myapp"`.
        app:
            an object implementing the [WSGI] or [ASGI] protocol.
        """
        if not prefix.startswith("/"):
            prefix = "/" + prefix

        self._children[prefix] = app

    def recipe(self, recipe: "Recipe"):
        """Apply a recipe.

        # Parameters
        recipe:
            a #::bocadillo.recipes#Recipe or #::bocadillo.recipes#RecipeBook
            to be applied to the application.

        # See Also
        - [Recipes](../guides/architecture/recipes.md)
        """
        recipe.apply(self)

    def add_error_handler(
        self, exception_cls: typing.Type[_E], handler: ErrorHandler
    ):
        """Register a new error handler.

        # Parameters
        exception_cls (exception class):
            The type of exception that should be handled.
        handler (callable):
            The actual error handler, which is called when an instance of
            `exception_cls` is caught.
            Should accept a request, response and exception parameters.
        """
        self.exception_middleware.add_exception_handler(exception_cls, handler)

    def error_handler(self, exception_cls: typing.Type[Exception]):
        """Register a new error handler (decorator syntax).

        # See Also
        - [add_error_handler](#add-error-handler)
        """

        def wrapper(handler):
            self.add_error_handler(exception_cls, handler)
            return handler

        return wrapper

    def add_middleware(self, middleware_cls, **kwargs):
        """Register a middleware class.

        # Parameters
        middleware_cls: a subclass of #::bocadillo.middleware#Middleware.

        # See Also
        - [Middleware](../guides/http/middleware.md)
        """
        self.exception_middleware.app = middleware_cls(
            self.exception_middleware.app, app=self, **kwargs
        )

    def add_asgi_middleware(self, middleware_cls, **kwargs):
        """Register an ASGI middleware class.

        # Parameters
        middleware_cls: a class that complies with the ASGI specification.

        # See Also
        - [ASGI middleware](../guides/agnostic/asgi-middleware.md)
        - [ASGI](https://asgi.readthedocs.io)
        """
        args = (self,) if issubclass(middleware_cls, ASGIMiddleware) else ()
        self.asgi = middleware_cls(self.asgi, *args, **kwargs)

    def on(self, event: str, handler: typing.Optional[EventHandler] = None):
        """Register an event handler.

        # Parameters
        event (str):
            Either `"startup"` (when the server boots) or `"shutdown"`
            (when the server stops).
        handler (callback, optional):
            The event handler. If not given, this should be used as a
            decorator.

        # Example

        ```python
        @app.on("startup")
        async def startup():
            pass

        async def shutdown():
            pass

        app.on("shutdown", shutdown)
        ```
        """
        if handler is None:

            def register(func):
                self._lifespan.add_event_handler(event, func)
                return func

            return register

        self._lifespan.add_event_handler(event, handler)
        return handler

    def dispatch_http(self, scope: Scope) -> ASGIAppInstance:
        async def asgi(receive: Receive, send: Send):
            req = Request(scope, receive)
            res = Response(req)

            res: Response = await self.server_error_middleware(req, res)
            await res(receive, send)
            # Re-raise the exception to allow the server to log the error
            # and for the test client to optionally re-raise it too.
            self.server_error_middleware.raise_if_exception()

        return asgi

    def dispatch_websocket(self, scope: Scope) -> ASGIAppInstance:
        async def asgi(receive: Receive, send: Send):
            await self.websocket_router(scope, receive, send)

        return asgi

    def dispatch(self, scope: Scope) -> ASGIAppInstance:
        path: str = scope["path"]

        # Return a sub-mounted extra app, if found
        for prefix, app in self._children.items():
            if not path.startswith(prefix):
                continue
            # Remove prefix from path so that the request is made according
            # to the mounted app's point of view.
            scope["path"] = path[len(prefix) :]
            try:
                return app(scope)
            except TypeError:
                return WSGIResponder(app, scope)

        if scope["type"] == "websocket":
            return self.dispatch_websocket(scope)

        assert scope["type"] == "http"
        return self.dispatch_http(scope)

    def __call__(self, scope: Scope) -> ASGIAppInstance:
        if scope["type"] == "lifespan":
            return self._lifespan(scope)
        return self.asgi(scope)



from os.path import exists

from whitenoise import WhiteNoise

from .compat import WSGIApp, empty_wsgi_app


def static(root: str, **kwargs) -> WSGIApp:
    """Return a WSGI app that serves static files under the given directory.

    Powered by [WhiteNoise](http://whitenoise.evans.io).

    [config-attrs]: http://whitenoise.evans.io/en/stable/base.html#configuration-attributes

    # Parameters
    root (str):
        the path to a directory from where static files should be served.
        If the directory does not exist, no files will be served.
    **kwargs (any):
        keyword arguments passed to the WhiteNoise constructor. See also [Configuration attributes (WhiteNoise docs)][config-attrs].

    # Returns
    app (callable): a WhiteNoise WSGI app.
    """
    if exists(root):
        kwargs["root"] = root
    return WhiteNoise(empty_wsgi_app(), **kwargs)
